"""
Parser for reading XPS (X-ray Photoelectron Spectroscopy) data from native
Specs Lab Prodigy SLE format, to be passed to mpes nxdl
(NeXus Definition Language) template.
"""

# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# pylint: disable=too-many-lines,too-many-instance-attributes

"""
Parser for reading XPS (X-ray Photoelectron Spectroscopy) metadata from
SPECS Lab Prodigy, to be passed to MPES nxdl (NeXus Definition Language)
template.
"""

import copy
import logging
import re
import sqlite3
import struct
import warnings
import xml.etree.ElementTree as ET
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Any

import numpy as np
import xarray as xr

from pynxtools_xps.reader_utils import (
    XPSMapper,
    construct_data_key,
    construct_entry_name,
    drop_unused_keys,
    re_map_keys,
    re_map_values,
)
from pynxtools_xps.value_mappers import (
    convert_energy_scan_mode,
    convert_energy_type,
    convert_measurement_method,
    convert_units,
    get_units_for_key,
)

logger = logging.getLogger(__name__)

UNITS: dict[str, str] = {
    "work_function": "eV",
    "excitation_energy": "eV",
    "iris_diameter": "mm",
    "step_size": "eV",
    "detector_voltage": "V",
    "dwell_time": "s",
    "raw_data/raw": "counts_per_second ",
    "polar_angle": "degree ",
    "azimuth_angle": "degree",
    "pass_energy": "eV",
    "start_energy": "eV",
    "emission_current": "A",
    "source_voltage": "V",
    "energy_calibration/energy": "eV",
    "transmission_function/kinetic_energy": "eV",
    "transmission_correction/transmission_function/kinetic_energy": "eV",
}


class SleMapperSpecs(XPSMapper):
    """
    Class for restructuring .sle data file from
    specs vendor into python dictionary.
    """

    config_file = "config_specs_sle.json"

    def __init__(self):
        self.parsers = [
            SleProdigyParserV1,
            SleProdigyParserV4,
        ]

        self.versions_map = {}
        for parser in self.parsers:
            supported_versions = parser.supported_versions
            for version in supported_versions:
                self.versions_map[version] = parser

        self.sql_connection = None

        super().__init__()

    def _select_parser(self):
        """
        Select the correct parser for the SLE file version.

        Returns
        -------
        Parser
            Specs SLE Parser.

        """
        version = self._get_sle_version()
        try:
            return self.versions_map[version]()
        except KeyError as exc:
            raise KeyError(
                f"Version f{version} of SPECS Prodigy is currently not supported."
            ) from exc

    def _get_sle_version(self):
        """Get the Prodigy SLE version from the file."""
        con = sqlite3.connect(self.sql_connection)
        cur = con.cursor()
        query = 'SELECT Value FROM Configuration WHERE Key=="Version"'
        cur.execute(query)
        version = cur.fetchall()[0][0]
        version = version.split(".")
        version = version[0] + "." + version[1].split("-")[0]
        return version

    def parse_file(self, file, **kwargs):
        """
        Parse the file using the parser that fits the Prodigy SLE version.
        Returns flat list of dictionaries containing one spectrum each.

        """
        self.sql_connection = file
        return super().parse_file(file, **kwargs)

    def construct_data(self):
        """Map SLE format to NXmpes-ready dict."""
        # pylint: disable=duplicate-code
        spectra = copy.deepcopy(self.raw_data)

        self._xps_dict["data"]: dict = {}

        for spectrum in spectra:
            self._update_xps_dict_with_spectrum(spectrum)

    def _update_xps_dict_with_spectrum(self, spectrum: dict[str, Any]):
        """
        Map one spectrum from raw data to NXmpes-ready dict.

        """
        # pylint: disable=too-many-locals,duplicate-code
        entry_parts = []
        for part in ["group_name", "spectrum_type"]:
            val = spectrum.get(part, None)
            if val:
                entry_parts += [val]

        entry = construct_entry_name(entry_parts)
        entry_parent = f"/ENTRY[{entry}]"

        for key, value in spectrum.items():
            if key.startswith("entry"):
                entry_parent = "/ENTRY[entry]"
                key = key.replace("entry/", "", 1)
            mpes_key = f"{entry_parent}/{key}"
            if "units" in key:
                value = convert_units(value)
            self._xps_dict[mpes_key] = value
            units = convert_units(get_units_for_key(key, UNITS))
            if units is not None:
                self._xps_dict[f"{mpes_key}/@units"] = units

        analyzer_path = f"{entry_parent}/instrument/electronanalyzer"

        self._xps_dict[f"{entry_parent}/electronanalyzer/name"] = spectrum["devices"][0]
        self._xps_dict[f"{entry_parent}/'source/name"] = spectrum["devices"][1]

        # Create keys for writing to data
        scan_key = construct_data_key(spectrum)

        energy = np.array(spectrum["data"]["x"])

        # If multiple spectra exist to entry, only create a new
        # xr.Dataset if the entry occurs for the first time.
        if entry not in self._xps_dict["data"]:
            self._xps_dict["data"][entry] = xr.Dataset()

        # Write averaged cycle data to 'data'.
        all_scan_data = [
            value
            for key, value in self._xps_dict["data"][entry].items()
            if scan_key.split("_")[0] in key
        ]

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            averaged_scans = np.mean(all_scan_data, axis=0)

        if averaged_scans.size == 1:
            # on first scan in cycle
            averaged_scans = spectrum["data"]["cps_calib"]

        if averaged_scans.shape == energy.shape:
            # TODO: fix this hotfix so that all data can be written

            self._xps_dict["data"][entry][scan_key.split("_")[0]] = xr.DataArray(
                data=averaged_scans,
                coords={"energy": energy},
            )

            # Write scan data to 'data'.
            self._xps_dict["data"][entry][scan_key] = xr.DataArray(
                data=spectrum["data"]["cps_calib"], coords={"energy": energy}
            )

            channels = [key for key in spectrum["data"] if "cps_ch_" in key]
            for channel in channels:
                ch_no = channel.rsplit("_")[-1]
                channel_key = f"{scan_key}_chan{ch_no}"
                # detector_data_key = (
                #     f"{path_map['detector']}/{detector_data_key_child}"
                #     f"_channels_Channel_{ch_no}/counts"
                # )
                cps = np.array(spectrum["data"][channel])

                # # Write raw data to detector.
                # self._xps_dict[detector_data_key] = spectrum["data"]["cps_calib"]
                # Write channel data to 'data'.
                self._xps_dict["data"][entry][channel_key] = xr.DataArray(
                    data=cps, coords={"energy": energy}
                )

            # Add unit for detector data
            detector_data_unit_key = f"{entry_parent}/detector/raw_data/raw/@units"

            detector_data_units = get_units_for_key("detector/raw_data/raw", UNITS)
            if detector_data_units is not None:
                self._xps_dict[detector_data_unit_key] = detector_data_units


class SleProdigyParser(ABC):
    """
    Generic parser without reading capabilities,
    to be used as template for implementing parsers for different versions.
    """

    def __init__(self):
        self.con = ""
        self.spectra = []
        self.xml = None
        self.sum_channels = False
        self.remove_align = True

        keys_map = {
            "Udet": "detector_voltage",
            "Comment": "comments",
            "ElectronEnergy": "start_energy",
            "SpectrumID": "spectrum_id",
            "EpassOrRR": "pass_energy",
            "EnergyType": "energy/@type",
            "Samples": "n_values",
            "Wf": "work_function",
            "Step": "step",
            "Ubias": "electron_bias",
            "DwellTime": "dwell_time",
            "NumScans": "total_scans",
            "LensMode": "lens_mode",
            "Timestamp": "time_stamp",
            "Entrance": "entrance_slit",
            "Exit": "exit_slit",
            "ScanMode": "energy_scan_mode",
            "VoltageRange": "voltage_energy_range",
        }

        spectrometer_setting_map = {
            "Coil Current [mA]": "coil_current [mA]",
            "Pre Defl Y [nU]": "pre_deflector_y_current [nU]",
            "Pre Defl X [nU]": "pre_deflector_x_current [nU]",
            "L1 [nU]": "lens1_voltage [nU]",
            "L2 [nU]": "lens2_voltage [nU]",
            "Focus Displacement 1 [nu]": "focus_displacement_current [nU]",
            "Detector Voltage [V]": "detector_voltage [V]",
            "Bias Voltage Electrons [V]": "bias_voltage_electrons [V]",
            "Bias Voltage Ions [V]": "bias_voltage_ions [V]",
        }

        source_setting_map = {
            "anode": "source_label",
            "uanode": "source_voltage",
            "iemission": "emission_current",
            "ihv": "source_high_voltage",
            "ufilament": "filament_voltage",
            "ifilament": "filament_current",
            "DeviceExcitationEnergy": "excitation_energy",
            "panode": "anode_power",
            "temperature": "source_temperature",
        }

        self.sql_metadata_map = {
            "EnergyType": "energy/@type",
            "EpassOrRR": "pass_energy",
            "Wf": "work_function",
            "Timestamp": "time_stamp",
            "Samples": "n_values",
            "ElectronEnergy": "start_energy",
            "Step": "step_size",
        }

        self.key_maps = [
            keys_map,
            spectrometer_setting_map,
            source_setting_map,
            self.sql_metadata_map,
        ]

        self.value_map = {
            "energy/@type": convert_energy_type,
            "excitation_energy": self._convert_excitation_energy,
            "time_stamp": self._convert_date_time,
            "energy_scan_mode": convert_energy_scan_mode,
        }

        self.keys_to_drop = [
            "Work Function",
        ]

        self.encoding = ["f", 4]

        self.measurement_types = ["XPS", "UPS", "ElectronSpectroscopy"]

    def initiate_file_connection(self, filepath):
        """Set the filename of the file to be opened."""
        sql_connection = filepath
        self.con = sqlite3.connect(sql_connection)

    def parse_file(self, filepath, **kwargs):
        """
        Parse the file's data and metadata into a flat list of dictionaries.


        Parameters
        ----------
        filename : str
            Filepath of the SLE file to be read.

        Returns
        -------
        self.spectra
            Flat list of dictionaries containing one spectrum each.

        """
        if "remove_align" in kwargs:
            self.remove_align = kwargs["remove_align"]

        try:
            self.sum_channels = kwargs["sum_channels"]
        except KeyError:
            self.sum_channels = False

        # initiate connection to sql file
        self.initiate_file_connection(filepath)

        # read and parse sle file
        self._get_xml_schedule()
        self.spectra = self._flatten_xml(self.xml)
        self._attach_node_ids()
        self._remove_empty_nodes()
        self._attach_device_protocols()
        self._get_spectrum_metadata_from_sql()
        self._check_encoding()

        self._append_scan_data()

        self._convert_to_common_format()
        self._close_con()

        if self.remove_align:
            self._remove_fixed_energies()

        self._remove_syntax()
        self._remove_snapshot()
        self._reindex_spectra()
        self._reindex_groups()

        return self.spectra

    def _get_xml_schedule(self):
        """Parse the schedule into an XML object."""
        cur = self.con.cursor()
        query = 'SELECT Value FROM Configuration WHERE Key="Schedule"'
        cur.execute(query)
        self.xml = ET.fromstring(cur.fetchall()[0][0])

    def _append_scan_data(self):
        """
        Get the signal data, convert to counts per seconds and get scan
        metadata (Scan no., loop no. and iteration no.) from each scan and
        attach to each spectrum.

        Returns
        -------
        None.

        """
        # pylint: disable=too-many-locals
        individual_scans = []
        scan_id = 0
        for spectrum in self.spectra:
            # copy node to new instance
            node_id = self._get_sql_node_id(spectrum["spectrum_id"])
            n_channels = self._check_energy_channels(node_id)
            raw_ids = self._get_raw_ids(node_id)
            n_scans = len(raw_ids)

            transmission_data = self._get_transmission(node_id)

            for scan_no in range(n_scans):
                scan = copy.copy(spectrum)

                scan["scan_id"] = scan_id
                # get signal data for each scan
                signal_data = self._get_one_scan(raw_ids[scan_no])

                # extract the individual channel data
                signal_data = self._separate_channels(signal_data, n_channels)

                # average channels if required
                if self.sum_channels:
                    signal_data = self._sum_channels(signal_data)

                # convert to counts per second
                signal_data_cps = self._convert_to_counts_per_sec(
                    signal_data, float(scan["dwell_time"])
                )

                # attach individual channel data to scan
                for ch_no, channel_data in enumerate(signal_data_cps):
                    scan[f"cps_ch_{ch_no}"] = list(channel_data)
                # no_of_scans_avg['scans'] = 1

                # scan['cps_calib'] = self._get_calibrated_data(spectrum)
                # """ This is wrong and needs to be corrected!!!"""
                scan["cps_calib"] = copy.copy(scan["cps_ch_0"])

                # Add transmission function
                scan["transmission_function/relative_intensity"] = np.array(
                    transmission_data
                )

                # add metadata including scan, loop no and datetime
                scan_metadata = self._get_scan_metadata(raw_ids[scan_no])
                for key, values in scan_metadata.items():
                    scan[key] = values

                individual_scans += [scan]
                scan_id += 1

        # update self.spectra with the scan data
        self.spectra = individual_scans

    def _get_transmission(self, node_id):
        """
        Get the transmission function data.

        Parameters
        ----------
        node_id : int
            Internal node ID of spectrum in SLE sql database.

        Returns
        -------
        transmission_data : array
            Array of TF values for the spectrum at node ID.
        """
        cur = self.con.cursor()
        query = (
            f'SELECT Data, SAMPLES, Ekin FROM TransmissionData WHERE Node="{node_id}"'
        )
        cur.execute(query)
        results = cur.fetchall()
        buffer = self.encoding[1]
        encoding = self.encoding[0]

        stream = []
        for result in results:
            length = result[1] * buffer
            data = result[0]
            for i in range(0, length, buffer):
                stream.append(struct.unpack(encoding, data[i : i + buffer])[0])

        return stream

    def _separate_channels(self, data, n_channels):
        """
        Separate energy channels.

        Parameters
        ----------
        data : list
            Array of measured daata .
        n_channels : int
            Number of channels to be summed.

        Returns
        -------
        list
            Summed data across n_channels.

        """

        n_points = int(len(data) / n_channels)
        return np.reshape(np.array(data), (n_channels, n_points))

    # """ NEED TO UPDATE THIS METHOD"""
    # def _get_calibrated_data(self, raw_data):
    #     """
    #
    #
    #     Parameters
    #     ----------
    #     raw_data : List
    #         DESCRIPTION.
    #
    #     Returns
    #     -------
    #     channel_dict : TYPE
    #         DESCRIPTION.
    #
    #     """
    #     mcd_num = int(raw_data["mcd_num"])
    #
    #     curves_per_scan = raw_data["curves_per_scan"]
    #     values_per_curve = raw_data["values_per_curve"]
    #     values_per_scan = int(curves_per_scan * values_per_curve)
    #     mcd_head = int(raw_data["mcd_head"])
    #     mcd_tail = int(raw_data["mcd_tail"])
    #     excitation_energy = raw_data["excitation_energy"]
    #     energy_scan_mode = raw_data["energy_scan_mode"]
    #     kinetic_energy = raw_data["kinetic_energy"]
    #     scan_delta = raw_data["scan_delta"]
    #     pass_energy = raw_data["pass_energy"]
    #     kinetic_energy_base = raw_data["kinetic_energy_base"]
    #     # Adding one unit to the binding_energy_upper is added as
    #     # electron comes out if energy is one unit higher
    #     binding_energy_upper = excitation_energy - \
    #         kinetic_energy + kinetic_energy_base + 1
    #
    #     mcd_energy_shifts = raw_data["mcd_shifts"]
    #     mcd_energy_offsets = []
    #     offset_ids = []
    #
    #     # consider offset values for detector with respect to
    #     # position at +16 which is usually large and positive value
    #     for mcd_shift in mcd_energy_shifts:
    #         mcd_energy_offset = (
    #             mcd_energy_shifts[-1] - mcd_shift) * pass_energy
    #         mcd_energy_offsets.append(mcd_energy_offset)
    #         offset_id = round(mcd_energy_offset / scan_delta)
    #         offset_ids.append(
    #             int(offset_id - 1 if offset_id > 0 else offset_id))
    #
    #     # Skiping entry without count data
    #     if not mcd_energy_offsets:
    #         continue
    #     mcd_energy_offsets = np.array(mcd_energy_offsets)
    #     # Putting energy of the last detector as a highest energy
    #     starting_eng_pnts = binding_energy_upper - mcd_energy_offsets
    #     ending_eng_pnts = (starting_eng_pnts
    #                        - values_per_scan * scan_delta)
    #
    #     channeltron_eng_axes = np.zeros((mcd_num, values_per_scan))
    #     for ind in np.arange(len(channeltron_eng_axes)):
    #         channeltron_eng_axes[ind, :] = \
    #             np.linspace(starting_eng_pnts[ind],
    #                         ending_eng_pnts[ind],
    #                         values_per_scan)
    #
    #     channeltron_eng_axes = np.round_(channeltron_eng_axes,
    #                                      decimals=8)
    #     # construct ultimate or incorporated energy axis from
    #     # lower to higher energy
    #     scans = list(raw_data["scans"].keys())
    #
    #     # Check whether array is empty or not
    #     if not scans:
    #         continue
    #     if not raw_data["scans"][scans[0]].any():
    #         continue
    #     # Sorting in descending order
    #     binding_energy = channeltron_eng_axes[-1, :]
    #
    #     self._xps_dict["data"][entry] = xr.Dataset()
    #
    #     for scan_nm in scans:
    #         channel_counts = np.zeros((mcd_num + 1,
    #                                    values_per_scan))
    #         # values for scan_nm corresponds to the data for each
    #         # "scan" in individual CountsSeq
    #         scan_counts = raw_data["scans"][scan_nm]
    #
    #         if energy_scan_mode == "fixed_analyzer_transmission":
    #             for row in np.arange(mcd_num):
    #
    #                 count_on_row = scan_counts[row::mcd_num]
    #                 # Reverse counts from lower to higher
    #                 # BE as in BE_eng_axis
    #                 count_on_row = \
    #                     count_on_row[mcd_head:-mcd_tail]
    #
    #                 channel_counts[row + 1, :] = count_on_row
    #                 channel_counts[0, :] += count_on_row
    #
    #                 # Storing detector's raw counts
    #                 self._xps_dict["data"][entry][f"{scan_nm}_chan_{row}"] = \
    #                     xr.DataArray(data=channel_counts[row + 1, :],
    #                                  coords={"BE": binding_energy})
    #
    #                 # Storing callibrated and after accumulated each scan counts
    #                 if row == mcd_num - 1:
    #                     self._xps_dict["data"][entry][scan_nm] = \
    #                         xr.DataArray(data=channel_counts[0, :],
    #                                      coords={"BE": binding_energy})
    #         else:
    #             for row in np.arange(mcd_num):
    #
    #                 start_id = offset_ids[row]
    #                 count_on_row = scan_counts[start_id::mcd_num]
    #                 count_on_row = count_on_row[0:values_per_scan]
    #                 channel_counts[row + 1, :] = count_on_row
    #
    #                 # shifting and adding all the curves.
    #                 channel_counts[0, :] += count_on_row
    #
    #                 # Storing detector's raw counts
    #                 self._xps_dict["data"][entry][f"{scan_nm}_chan{row}"] = \
    #                     xr.DataArray(data=channel_counts[row + 1, :],
    #                                  coords={"BE": binding_energy})
    #
    #                 # Storing callibrated and after accumulated each scan counts
    #                 if row == mcd_num - 1:
    #                     self._xps_dict["data"][entry][scan_nm] = \
    #                         xr.DataArray(data=channel_counts[0, :],
    #                                      coords={"BE": binding_energy})
    #
    #     # Skiping entry without count data
    #     if not mcd_energy_offsets:
    #         continue
    #     mcd_energy_offsets = np.array(mcd_energy_offsets)
    #     # Putting energy of the last detector as a highest energy
    #     starting_eng_pnts = binding_energy_upper - mcd_energy_offsets
    #     ending_eng_pnts = (starting_eng_pnts
    #                        - values_per_scan * scan_delta)
    #
    #     channeltron_eng_axes = np.zeros((mcd_num, values_per_scan))
    #     for ind in np.arange(len(channeltron_eng_axes)):
    #         channeltron_eng_axes[ind, :] = \
    #             np.linspace(starting_eng_pnts[ind],
    #                         ending_eng_pnts[ind],
    #                         values_per_scan)
    #
    #     channeltron_eng_axes = np.round_(channeltron_eng_axes,
    #                                      decimals=8)
    #     # construct ultimate or incorporated energy axis from
    #     # lower to higher energy
    #     scans = list(raw_data["scans"].keys())
    #
    #     # Check whether array is empty or not
    #     if not scans:
    #         continue
    #     if not raw_data["scans"][scans[0]].any():
    #         continue
    #     # Sorting in descending order
    #     binding_energy = channeltron_eng_axes[-1, :]
    #
    #     self._xps_dict["data"][entry] = xr.Dataset()
    #
    #     for scan_nm in scans:
    #         channel_counts = np.zeros((mcd_num + 1,
    #                                    values_per_scan))
    #         # values for scan_nm corresponds to the data for each
    #         # "scan" in individual CountsSeq
    #         scan_counts = raw_data["scans"][scan_nm]
    #
    #         if energy_scan_mode == "fixed_analyzer_transmission":
    #             for row in np.arange(mcd_num):
    #
    #                 count_on_row = scan_counts[row::mcd_num]
    #                 # Reverse counts from lower to higher
    #                 # BE as in BE_eng_axis
    #                 count_on_row = \
    #                     count_on_row[mcd_head:-mcd_tail]
    #
    #                 channel_counts[row + 1, :] = count_on_row
    #                 channel_counts[0, :] += count_on_row
    #
    #                 # Storing detector's raw counts
    #                 self._xps_dict["data"][entry][f"{scan_nm}_chan_{row}"] = \
    #                     xr.DataArray(data=channel_counts[row + 1, :],
    #                                  coords={"BE": binding_energy})
    #
    #                 # Storing callibrated and after accumulated each scan counts
    #                 if row == mcd_num - 1:
    #                     self._xps_dict["data"][entry][scan_nm] = \
    #                         xr.DataArray(data=channel_counts[0, :],
    #                                      coords={"BE": binding_energy})
    #         else:
    #             for row in np.arange(mcd_num):
    #
    #                 start_id = offset_ids[row]
    #                 count_on_row = scan_counts[start_id::mcd_num]
    #                 count_on_row = count_on_row[0:values_per_scan]
    #                 channel_counts[row + 1, :] = count_on_row
    #
    #                 # shifting and adding all the curves.
    #                 channel_counts[0, :] += count_on_row
    #
    #                 # Storing detector's raw counts
    #                 self._xps_dict["data"][entry][f"{scan_nm}_chan{row}"] = \
    #                     xr.DataArray(data=channel_counts[row + 1, :],
    #                                  coords={"BE": binding_energy})
    #
    #                 # Storing callibrated and after accumulated each scan counts
    #                 if row == mcd_num - 1:
    #                     self._xps_dict["data"][entry][scan_nm] = \
    #                         xr.DataArray(data=channel_counts[0, :],
    #                                      coords={"BE": binding_energy})
    #
    #     return channel_dict

    def _check_energy_channels(self, node_id):
        """
        Get the number of separate energy channels for the spectrum.

        This checks to see if the spectrum was saved with separated energy
        channels.

        Parameters
        ----------
        node_id : int
            Internal node ID of spectrum in SLE sql database.

        Returns
        -------
        n_channels : int
            Number of separate energy channels for the spectrum at node ID.
        """
        cur = self.con.cursor()
        query = f'SELECT EnergyChns FROM Spectrum WHERE Node="{node_id}"'
        cur.execute(query)
        result = cur.fetchall()
        if len(result) != 0:
            n_channels = result[0][0]
        return n_channels

    def _get_raw_ids(self, node_id):
        """
        Get the raw IDs from SQL.

        There is one raw_id for each individual scan when scans were not
        already averaged in the sle file.
        To know which rows in the detector data table belong to which scans,
        one needs to first get the raw_id from the RawData table.

        Parameters
        ----------
        node_id : int
            Internal node ID of spectrum in SLE sql database.

        Returns
        -------
        list
            List of raw IDs for the given note ID.

        """
        cur = self.con.cursor()
        query = f'SELECT RawId FROM RawData WHERE Node="{node_id}"'
        cur.execute(query)

        return [i[0] for i in cur.fetchall()]

    def _check_number_of_scans(self, node_id):
        """
        Get the number of separate scans for the spectrum.

        Parameters
        ----------
        node_id : int
            Internal node ID of spectrum in SLE sql database.

        Returns
        -------
        int
            Number of separate scans for the spectrum.

        """
        cur = self.con.cursor()
        query = f'SELECT RawId FROM RawData WHERE Node="{node_id}"'
        cur.execute(query)
        return len(cur.fetchall())

    def _get_detector_data(self, node_id):
        """
        Get the detector data from sle file.

        The detector data is stored in the SQLite database as a blob.
        To know which blobs belong to which scans, one needs to first get the
        raw_id from the RawData table.

        Parameters
        ----------
        node_id : int
            Internal node ID of spectrum in SLE sql database.

        Returns
        -------
        detector_data : list
            List of lists with measured data.

        """
        cur = self.con.cursor()
        query = f'SELECT RawID FROM RawData WHERE Node="{node_id}"'
        cur.execute(query)
        raw_ids = [i[0] for i in cur.fetchall()]
        detector_data = []
        if len(raw_ids) > 1:
            for raw_id in raw_ids:
                detector_data += [self._get_one_scan(raw_id)]
        else:
            raw_id = raw_ids[0]
            detector_data = self._get_one_scan(raw_id)

        return detector_data

    def _attach_device_protocols(self):
        """
        Get the device protocol for each node and add the paramaters of
        the Phoibos to the spectra table. Occassionally these are not
        recorded, if this is the case just skip the group.

        Returns
        -------
        None.

        """
        # iterate through each spectrum
        for spectrum in self.spectra:
            # conver the xml xps id to the node ID and get the device protocol
            cur = self.con.cursor()
            protocol_node_id = self._get_sql_node_id(spectrum["device_group_id"])
            query = (
                f'SELECT Protocol FROM DeviceProtocol WHERE Node="{protocol_node_id}"'
            )
            result = cur.execute(query).fetchone()

            # if a record was accessed then parse, if not skip
            if result:
                protocol = ET.fromstring(result[0])
                protocal_params = self._get_one_device_protocol(protocol)
                spectrum.update(protocal_params)

    def _get_one_device_protocol(self, protocol):
        """
         Get all parameters for one device protocol

        Parameters
        ----------
        protocol : xml.etree.ElementTree.Element
            One device protocol.

        Returns
        -------
        protocal_params : dict
            All parameters given in the device protocol.

        """
        protocal_params = {}
        for device in protocol.iter("Command"):
            if "Phoibos" in device.attrib["UniqueDeviceName"]:
                # iterate through the parameters and add to spectrum
                # dict
                for parameter in device.iter("Parameter"):
                    if parameter.attrib["type"] == "double":
                        param_text = float(parameter.text)
                    else:
                        param_text = parameter.text
                    protocal_params[parameter.attrib["name"]] = param_text
            elif "XRC1000" in device.attrib["UniqueDeviceName"]:
                for parameter in device.iter("Parameter"):
                    if parameter.attrib["type"] == "double":
                        param_text = float(parameter.text)
                    else:
                        param_text = parameter.text
                    protocal_params[parameter.attrib["name"]] = param_text

        return protocal_params

    def _get_one_scan(self, raw_id):
        """
        Get the detector data for a single scan and convert it to float.

        The detector data is stored in the SQLite database as a blob.
        This function decodes the blob into python float. The blob can be
        enoded as float or double in the SQLite table.

        Parameters
        ----------
        raw_id : int
            Raw ID of the single scan.

        Returns
        -------
        stream : list
            List with measured data.

        """
        cur = self.con.cursor()
        query = f'SELECT Data, ChunkSize FROM CountRateData WHERE RawId="{raw_id}"'
        cur.execute(query)
        results = cur.fetchall()
        buffer = self.encoding[1]
        encoding = self.encoding[0]
        stream = []
        for result in results:
            length = result[1] * buffer
            data = result[0]
            for i in range(0, length, buffer):
                stream.append(struct.unpack(encoding, data[i : i + buffer])[0])
        return stream

    def _parse_external_channels(self, channel):
        """
        Parse additional external channels.

        Parameters
        ----------
        channel : int
            DESCRIPTION.

        Returns
        -------
        None.

        """
        if len(channel) != 0:
            pass

    def _get_spectrum_metadata_from_sql(self):
        """
        Get the metadata stored in the SQLite Spectrum table

        Returns
        -------
        None.

        """
        for spectrum in self.spectra:
            node_id = self._get_sql_node_id(spectrum["spectrum_id"])
            cur = self.con.cursor()
            query = f'SELECT * FROM Spectrum WHERE Node="{node_id}"'
            cur.execute(query)
            results = cur.fetchall()
            if len(results) != 0:
                results = results[0]

            column_names = self._get_column_names("Spectrum")
            combined = {
                k: v
                for k, v in dict(zip(column_names, results)).items()
                if k in self.sql_metadata_map
            }
            combined = copy.copy(combined)
            if "EnergyType" not in combined.keys():
                combined["EnergyType"] = "Binding"
            for key, value in combined.items():
                spectrum[key] = value

            query = f'SELECT Data FROM NodeData WHERE Node="{node_id}"'
            cur.execute(query)
            results = ET.fromstring(cur.fetchall()[0][0])
            for i in results.iter("AnalyzerSpectrumParameters"):
                spectrum["work_function"] = i.attrib["Workfunction"]
                spectrum["step_size"] = float(i.attrib["ScanDelta"])

    def _get_scan_metadata(self, raw_id):
        """
        Get the scan and the loop/iteration number of each spectrum scan
        and the datetime it was taken from the RawData table.

        Parameters
        ----------
        raw_id : int
            raw id of the scan in the RawData table.

        Returns
        -------
        scan_meta : dict
            dictionary containing scan metadata.

        """
        # get string Trace from RawData
        cur = self.con.cursor()
        query = f'SELECT ScanDate, Trace FROM RawData WHERE RawID="{raw_id}"'
        result = cur.execute(query).fetchone()
        # process metadata into a dictionary
        scan_meta = {}
        scan_meta["time_stamp_trace"] = result[0]
        scan_meta.update(self._process_trace(result[1]))

        return scan_meta

    def _process_trace(self, trace):
        """
        Parse Trace string to determine the Scan, loop and iteration for the
        given trace.

        Parameters
        ----------
        trace : str
            string to be parsed.

        Returns
        -------
        trace_dict : dict
            dictionary containing scan loop and iteration params
        """
        trace_dict = {}
        loop = re.findall(r"Loop=([0-9]+)u", trace)
        if len(loop) != 0:
            trace_dict["loop_no"] = loop[0]

        scan = re.findall(r"Scan [\[Idx\] ]+=([0-9]+)u", trace)
        if len(scan) != 0:
            trace_dict["scan_no"] = scan[0]

        ramp = re.findall(r"Ramping Iteration [\[Idx\] ]+=([0-9]+)u", trace)
        if len(ramp) != 0:
            trace_dict["iteration_no"] = ramp[0]

        return trace_dict

    def _convert_to_counts_per_sec(self, signal_data, dwell_time):
        """
        Convert signal data given in counts to counts per second.

        Parameters
        ----------
        signal_data : list
            2D array of floats representing counts
            Shape: (n_channel, n_value)
        dwell_time : float
            value of dwell_time per scan.

        Returns
        -------
        cps : array
            2D array of values converted to counts per second.
            Shape: (n_channel, n_value)

        """
        cps = signal_data / dwell_time
        return cps

    def _get_sql_node_id(self, xml_id):
        """
        Get the SQL internal ID for the NodeID taken from XML.

        Sometimes the NodeID used in XML does not eaxtly map to the IDs for
        Spectra in the SQL tables. To fix this, there is a node mapping.

        Parameters
        ----------
        xml_id : int
            ID in the XML schedule.

        Returns
        -------
        node_id : int
            ID in the SQL tables.

        """
        cur = self.con.cursor()
        query = f'SELECT Node FROM NodeMapping WHERE InternalID="{xml_id}"'
        cur.execute(query)
        node_id = cur.fetchall()[0][0]
        return node_id

    def _attach_node_ids(self):
        """
        Attach the node_id to each spectrum in the spectra list.

        Returns
        -------
        None.

        """
        for spectrum in self.spectra:
            xml_id = spectrum["spectrum_id"]
            node_id = self._get_sql_node_id(xml_id)
            spectrum["node_id"] = node_id

    def _remove_empty_nodes(self):
        """
        Remove entries from spectra list that have no spectrum in SQLite.

        Returns
        -------
        None.
        """
        for j in reversed(list(enumerate(self.spectra))):
            idx = j[0]
            spectrum = j[1]
            node_id = spectrum["node_id"]
            cur = self.con.cursor()
            query = f'SELECT Node FROM Spectrum WHERE Node="{node_id}"'
            cur.execute(query)
            result = cur.fetchall()
            if len(result) == 0:
                del self.spectra[idx]

    def _get_energy_data(self, spectrum):
        """
        Create an array of x values.

        Parameters
        ----------
        spectrum : dict
            Dictionary with spectrum data and metadata.

        Returns
        -------
        x : list
            List of uniformly separated energy values.

        """
        if spectrum["energy/@type"] == "binding":
            start = spectrum["start_energy"]
            step = spectrum["step_size"]
            points = spectrum["n_values"]
            energy = [start - i * step for i in range(points)]
        elif spectrum["energy/@type"] == "kinetic":
            start = spectrum["start_energy"]
            step = spectrum["step_size"]
            points = spectrum["n_values"]
            energy = [start + i * step for i in range(points)]
        return np.array(energy)

    def _get_table_names(self):
        """
        Get a list of table names in the current database file.

        Returns
        -------
        data : list
            List of spectrum names.

        """
        cur = self.con.cursor()
        cur.execute('SELECT name FROM sqlite_master WHERE type= "table"')
        data = [i[0] for i in cur.fetchall()]
        return data

    def _get_column_names(self, table_name):
        """
        Get the names of the columns in the table.

        Parameters
        ----------
        table_name : str
            Name of SQL table.

        Returns
        -------
        names : list
            List of descriptions.

        """
        cur = self.con.cursor()
        cur.execute(f"SELECT * FROM {table_name}")
        names = [description[0] for description in cur.description]
        return names

    def _close_con(self):
        """
        Close the database connection.

        Returns
        -------
        None.

        """
        self.con.close()

    def _convert_excitation_energy(self, excitation_energy):
        """
        Convert the excitation_energy to a float.

        """
        return float(excitation_energy)

    def _convert_date_time(self, timestamp):
        """
        Convert the native time format to the one we decide to use.
        Returns datetime string in the format '%Y-%b-%d %H:%M:%S.%f'.

        """
        date_time = datetime.strptime(timestamp, "%Y-%b-%d %H:%M:%S.%f")
        date_time = datetime.strftime(date_time, "%Y-%m-%d %H:%M:%S.%f")
        return date_time

    def _sum_channels(self, data):
        """
        Sum together energy channels.

        Parameters
        ----------
        data : list
            Array of measured daata .
        n : int
            Number of channels to be summed.

        Returns
        -------
        list
            Summed data across n_channels.

        """
        summed = np.sum(data, axis=0)
        return np.reshape(summed, (1, -1))

    def _check_encoding(self):
        """
        Check whether the binary data should be decoded float or double.

        Returns
        -------
        None.

        """
        cur = self.con.cursor()
        query = "SELECT LENGTH(Data),ChunkSize FROM CountRateData LIMIT 1"
        cur.execute(query)
        data, chunksize = cur.fetchall()[0]

        encodings_map = {
            "double": ["d", 8],
            "float": ["f", 4],
        }

        if data / chunksize == 4:
            self.encoding = encodings_map["float"]
        elif data / chunksize == 8:
            self.encoding = encodings_map["double"]
        else:
            logger.error("This binary encoding is not supported.")

    @abstractmethod
    def _flatten_xml(self, xml):
        """
        Flatten the nested XML structure, keeping only the needed metadata.

        This method has to be implemented in the inherited parsers.

        Parameters
        ----------
        xml : xml.etree.ElementTree
            XML schedule of the experiment.

        Returns
        -------
        collect : list
            List of dictionary with spectra metadata.

        """
        # pylint: disable=too-many-nested-blocks
        collect = []

        return collect

    def _reindex_spectra(self):
        """Re-number the spectrum_id."""
        for idx, spectrum in enumerate(self.spectra):
            spectrum["spectrum_id"] = idx

    def _reindex_groups(self):
        """Re-number the group_id."""
        group_ids = list({spec["group_id"] for spec in self.spectra})
        for idx, group_id in enumerate(group_ids):
            for spec in self.spectra:
                if int(spec["group_id"]) == int(group_id):
                    spec["group_id"] = copy.copy(idx)

    def _convert_to_common_format(self):
        """
        Reformat spectra into the format needed for the Mapper object
        """
        maps = {}
        for key_map in self.key_maps:
            maps.update(key_map)
        for spec in self.spectra:
            re_map_keys(spec, maps)
            re_map_values(spec, self.value_map)
            drop_unused_keys(spec, self.keys_to_drop)
            spec["data"] = {}
            spec["data"]["x"] = self._get_energy_data(spec)

            channels = [
                key
                for key in spec
                if any(name in key for name in ["cps_ch_", "cps_calib"])
            ]

            for channel_key in channels:
                spec["data"][channel_key] = np.array(spec[channel_key])
            for channel_key in channels:
                spec.pop(channel_key)

            spec["energy/@units"] = "eV"
            spec["intensity/@units"] = "counts_per_second"

            # Add energy axis for TF data.
            if spec["energy/@type"] == "binding":
                tf_energy = np.array(
                    [spec["excitation_energy"] - x for x in spec["data"]["x"]]
                )
            elif spec["energy/@type"] == "kinetic":
                tf_energy = spec["data"]["x"]

            spec["transmission_function/kinetic_energy"] = tf_energy

    def _remove_fixed_energies(self):
        """
        Remove spectra measured with the scan mode FixedEnergies.
        """
        self.spectra = [
            spec for spec in self.spectra if spec["energy_scan_mode"] != "fixed_energy"
        ]

    def _remove_syntax(self):
        """
        Remove the extra syntax in the group name.
        """
        for spectrum in self.spectra:
            new_name = spectrum["group_name"].split("#", 1)[0]
            new_name = new_name.rstrip(", ")
            spectrum["group_name"] = new_name

    def _remove_snapshot(self):
        """
        Remove spectra required in Snapshot mode.
        """
        self.spectra = [
            spec for spec in self.spectra if "snapshot" not in spec["energy_scan_mode"]
        ]

    def get_sle_version(self):
        """
        Get the Prodigy SLE version from the file.

        Returns
        -------
        version : str
            Prodigy SLE version of SLE file.

        """
        cur = self.con.cursor()
        query = 'SELECT Value FROM Configuration WHERE Key=="Version"'
        cur.execute(query)
        version = cur.fetchall()[0][0]
        return version


class SleProdigyParserV1(SleProdigyParser):
    """
    Parser for SLE version 1.
    """

    supported_versions = ["1.2", "1.8", "1.9", "1.10", "1.11", "1.12", "1.13"]

    def _flatten_xml(self, xml):
        """
        Flatten the nested XML structure, keeping only the needed metadata.

        Parameters
        ----------
        xml : xml.etree.ElementTree
            XML schedule of the experiment.

        Returns
        -------
        collect : list
            List of dictionary with spectra metadata.

        """
        collect = []
        for measurement_type in self.measurement_types:
            for group in xml.iter(measurement_type):
                data = {}
                data["analysis_method"] = convert_measurement_method(measurement_type)

                data["devices"] = []

                for device in group.iter("DeviceCommand"):
                    settings = {}
                    for param in device.iter("Parameter"):
                        settings[param.attrib["name"]] = param.text
                        data.update(copy.copy(settings))

                    data["devices"] += [device.attrib["DeviceType"]]

                    # data['devices'] += [{'device_type' : j.attrib['DeviceType'],
                    #                     'settings':settings}]
                for spectrum_group in group.iter("SpectrumGroup"):
                    settings = self._get_group_metadata(spectrum_group)
                    data.update(copy.copy(settings))
                    collect += [copy.copy(data)]
        return collect

    def _get_group_metadata(self, spectrum_group):
        """
        Iteratively retrieve metadata for one spectrum group.

        Parameters
        ----------
        spectrum_group: xml.etree.ElementTree.Element
            XML element containing one spectrum group.

        Returns
        -------
        settings: dict
            Dictionary containing all metadata for
            the spectrum group.

        """
        settings = {}
        settings["group_name"] = spectrum_group.attrib["Name"]
        settings["group_id"] = spectrum_group.attrib["ID"]
        for comm_settings in spectrum_group.iter("CommonSpectrumSettings"):
            common_spectrum_settings = self._extract_comm_settings(comm_settings)
            settings.update(copy.copy(common_spectrum_settings))

        for spectrum in spectrum_group.iter("Spectrum"):
            spectrum_settings = self._get_spectrum_metadata(spectrum)
            settings.update(copy.copy(spectrum_settings))

        return settings

    def _extract_comm_settings(self, comm_settings):
        """
        Iteratively retrieve metadata for common settings of one spectrum group.

        Parameters
        ----------
        spectrum_group: xml.etree.ElementTree.Element
            XML element containing common settings for one spectrum group.

        Returns
        -------
        settings: dict
            Dictionary containing all common metadata for
            the spectrum group.

        """
        common_spectrum_settings = {}
        for setting in comm_settings.iter():
            if setting.tag == "ScanMode":
                energy_scan_mode = self.energy_scan_mode_map[setting.attrib["Name"]]
                common_spectrum_settings[setting.tag] = energy_scan_mode
            elif setting.tag == "SlitInfo":
                for key, val in setting.attrib.items():
                    common_spectrum_settings[key] = val
            elif setting.tag == "Lens":
                voltage_range = setting.attrib["VoltageRange"]
                split_text = re.split(r"([A-Z])", voltage_range, 1)
                val = split_text[0]
                unit = "".join(split_text[1:])
                common_spectrum_settings["voltage_range"] = float(val)
                common_spectrum_settings["voltage_range/@units"] = unit
            elif setting.tag == "EnergyChannelCalibration":
                common_spectrum_settings["calibration_file/dir"] = setting.attrib["Dir"]
                common_spectrum_settings["calibration_file/path"] = setting.attrib[
                    "File"
                ]
            elif setting.tag == "Transmission":
                common_spectrum_settings["transmission_function/file"] = setting.attrib[
                    "File"
                ]
            elif setting.tag == "Iris":
                common_spectrum_settings["iris_diameter"] = float(
                    setting.attrib["Diameter"]
                )
        return common_spectrum_settings

    def _get_spectrum_metadata(self, spectrum):
        """
        Iteratively retrieve metadata for one spectrum.

        Parameters
        ----------
        spectrum: xml.etree.ElementTree.Element
            XML element containing one spectrum.

        Returns
        -------
        spectrum_ settings: dict
            Dictionary containing all metadata for
            the spectrum.

        """
        spectrum_settings = {}

        spectrum_settings["spectrum_id"] = spectrum.attrib["ID"]
        spectrum_settings["spectrum_type"] = spectrum.attrib["Name"]
        for setting in spectrum.iter("FixedEnergiesSettings"):
            spectrum_settings["dwell_time"] = float(setting.attrib["DwellTime"])
            spectrum_settings["start_energy"] = float(copy.copy(setting.attrib["Ebin"]))
            spectrum_settings["pass_energy"] = float(setting.attrib["Epass"])
            spectrum_settings["lens_mode"] = setting.attrib["LensMode"]
            spectrum_settings["total_scans"] = int(setting.attrib["NumScans"])
            spectrum_settings["n_values"] = int(setting.attrib["NumValues"])
            spectrum_settings["end_energy"] = float(setting.attrib["End"])
            spectrum_settings["excitation_energy"] = float(setting.attrib["Eexc"])
            spectrum_settings["step_size"] = (
                spectrum_settings["start_energy"] - spectrum_settings["end_energy"]
            ) / (spectrum_settings["n_values"] - 1)
        for setting in spectrum.iter("FixedAnalyzerTransmissionSettings"):
            spectrum_settings["dwell_time"] = float(setting.attrib["DwellTime"])
            spectrum_settings["start_energy"] = float(copy.copy(setting.attrib["Ebin"]))
            spectrum_settings["pass_energy"] = float(setting.attrib["Epass"])
            spectrum_settings["lens_mode"] = setting.attrib["LensMode"]
            spectrum_settings["total_scans"] = setting.attrib["NumScans"]
            spectrum_settings["n_values"] = int(setting.attrib["NumValues"])
            spectrum_settings["end_energy"] = float(setting.attrib["End"])
            spectrum_settings["scans"] = int(setting.attrib["NumScans"])
            spectrum_settings["excitation_energy"] = float(setting.attrib["Eexc"])
            spectrum_settings["step_size"] = (
                spectrum_settings["start_energy"] - spectrum_settings["end_energy"]
            ) / (spectrum_settings["n_values"] - 1)
        return spectrum_settings


class SleProdigyParserV4(SleProdigyParser):
    """
    Parser for SLE version 4.
    """

    supported_versions = [
        "4.63",
        "4.64",
        "4.65",
        "4.66",
        "4.67",
        "4.68",
        "4.69",
        "4.70",
        "4.71",
        "4.72",
        "4.73",
    ]

    def _flatten_xml(self, xml):
        """
        Flatten the nested XML structure, keeping only the needed metadata.

        Parameters
        ----------
        xml : xml.etree.ElementTree
            XML schedule of the experiment.

        Returns
        -------
        collect : list
            List of dictionary with spectra metadata.

        """
        collect = []
        for measurement_type in self.measurement_types:
            for group in xml.iter(measurement_type):
                data = {}
                data["analysis_method"] = convert_measurement_method(measurement_type)

                data["devices"] = []
                data["device_group_id"] = group.attrib["ID"]

                for device in group.iter("DeviceCommand"):
                    settings = {}
                    for param in device.iter("Parameter"):
                        settings[param.attrib["name"]] = param.text
                        data.update(copy.copy(settings))

                    data["devices"] += [device.attrib["DeviceType"]]

                for spectrum_group in group.iter("SpectrumGroup"):
                    settings = self._get_group_metadata(spectrum_group)
                    data.update(copy.copy(settings))
                    collect += [copy.copy(data)]
        return collect

    def _get_group_metadata(self, spectrum_group):
        """
        Iteratively retrieve metadata for one spectrum group.

        Parameters
        ----------
        spectrum_group: xml.etree.ElementTree.Element
            XML element containing one spectrum group.

        Returns
        -------
        settings: dict
            Dictionary containing all metadata for
            the spectrum group.

        """
        settings = {}
        settings["group_name"] = spectrum_group.attrib["Name"]
        settings["group_id"] = spectrum_group.attrib["ID"]
        for comm_settings in spectrum_group.iter("CommonSpectrumSettings"):
            common_spectrum_settings = self._extract_comm_settings(comm_settings)
            settings.update(copy.copy(common_spectrum_settings))

        for spectrum in spectrum_group.iter("Spectrum"):
            spectrum_settings = self._get_spectrum_metadata(spectrum)
            settings.update(copy.copy(spectrum_settings))

        return settings

    def _extract_comm_settings(self, comm_settings):
        """
        Iteratively retrieve metadata for common settings of one spectrum group.

        Parameters
        ----------
        spectrum_group: xml.etree.ElementTree.Element
            XML element containing common settings for one spectrum group.

        Returns
        -------
        settings: dict
            Dictionary containing all common metadata for
            the spectrum group.

        """
        common_spectrum_settings = {}
        for setting in comm_settings.iter():
            if setting.tag == "ScanMode":
                energy_scan_mode = convert_energy_scan_mode(setting.attrib["Name"])
                common_spectrum_settings[setting.tag] = energy_scan_mode
            elif setting.tag == "SlitInfo":
                for key, val in setting.attrib.items():
                    common_spectrum_settings[key] = val
            elif setting.tag == "Lens":
                voltage_range = setting.attrib["VoltageRange"]
                match = re.match(r"(\d+\.?\d*)([a-zA-Z]+)", voltage_range)
                if match:
                    value, unit = match.groups()
                else:
                    value, unit = None, None
                common_spectrum_settings["voltage_energy_range"] = float(value)
                common_spectrum_settings["voltage_energy_range/@units"] = unit
            elif setting.tag == "EnergyChannelCalibration":
                common_spectrum_settings["calibration_file/dir"] = setting.attrib["Dir"]
                common_spectrum_settings["calibration_file/path"] = setting.attrib[
                    "File"
                ]
            elif setting.tag == "Transmission":
                common_spectrum_settings["transmission_function/file"] = setting.attrib[
                    "File"
                ]
            elif setting.tag == "Iris":
                common_spectrum_settings["iris_diameter"] = float(
                    setting.attrib["Diameter"]
                )
        return common_spectrum_settings

    def _get_spectrum_metadata(self, spectrum):
        """
        Iteratively retrieve metadata for one spectrum.

        Parameters
        ----------
        spectrum: xml.etree.ElementTree.Element
            XML element containing one spectrum.

        Returns
        -------
        spectrum_ settings: dict
            Dictionary containing all metadata for
            the spectrum.

        """
        spectrum_settings = {}

        spectrum_settings["spectrum_id"] = spectrum.attrib["ID"]
        spectrum_settings["spectrum_type"] = spectrum.attrib["Name"]
        for comment in spectrum.iter("Comment"):
            spectrum_settings["spectrum_comment"] = comment.text

        for setting in spectrum.iter("FixedEnergiesSettings"):
            spectrum_settings["dwell_time"] = float(setting.attrib["DwellTime"])
            spectrum_settings["start_energy"] = float(copy.copy(setting.attrib["Ebin"]))
            spectrum_settings["pass_energy"] = float(setting.attrib["Epass"])
            spectrum_settings["lens_mode"] = setting.attrib["LensMode"]
            spectrum_settings["total_scans"] = int(setting.attrib["NumScans"])
            spectrum_settings["n_values"] = int(setting.attrib["NumValues"])
        for setting in spectrum.iter("FixedAnalyzerTransmissionSettings"):
            spectrum_settings["dwell_time"] = float(setting.attrib["DwellTime"])
            spectrum_settings["start_energy"] = float(copy.copy(setting.attrib["Ebin"]))
            spectrum_settings["pass_energy"] = float(setting.attrib["Epass"])
            spectrum_settings["lens_mode"] = setting.attrib["LensMode"]
            spectrum_settings["total_scans"] = setting.attrib["NumScans"]
            spectrum_settings["n_values"] = int(setting.attrib["NumValues"])
            spectrum_settings["end_energy"] = float(setting.attrib["End"])
            spectrum_settings["scans"] = int(setting.attrib["NumScans"])
            spectrum_settings["step_size"] = (
                spectrum_settings["start_energy"] - spectrum_settings["end_energy"]
            ) / (spectrum_settings["n_values"] - 1)
        return spectrum_settings



# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# pylint: disable=too-many-lines
"""
Parser for reading XPS (X-ray Photoelectron Spectroscopy) metadata from
VAMAS standard, to be passed to MPES nxdl (NeXus Definition Language)
template.
"""

import datetime
import logging
import warnings
from copy import deepcopy
from itertools import groupby
from pathlib import Path
from typing import Any, Union

import numpy as np
import xarray as xr

from pynxtools_xps.reader_utils import (
    XPSMapper,
    check_for_allowed_in_list,
    construct_data_key,
    construct_entry_name,
    convert_pascal_to_snake,
    drop_unused_keys,
    get_minimal_step,
    re_map_keys,
    re_map_values,
    update_dict_without_overwrite,
)
from pynxtools_xps.value_mappers import (
    convert_energy_scan_mode,
    convert_energy_type,
    convert_measurement_method,
    convert_units,
    get_units_for_key,
)
from pynxtools_xps.vms.vamas_comment_handler import handle_comments
from pynxtools_xps.vms.vamas_data_model import (
    ExpVariable,
    OrdinateValue,
    VamasAdditionalParam,
    VamasBlock,
    VamasHeader,
)

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

EXP_MODES = [
    "MAP",
    "MAPDP",
    "MAPSV",
    "MAPSVDP",
    "NORM",
    "SDP",
    "SDPSC",
    "SEM",
]

ALLOWED_TECHNIQUES = [
    "AES",
    "AES diff",
    "AES dir",
    "EDX",
    "ELS",
    "FABMS",
    "FABMS energy spec",
    "ISS",
    "SIMS",
    "SIMS energy spec",
    "SNMS",
    "SNMS energy spec",
    "UPS",
    "XPS",
    "XRF",
]

UNITS: dict = {
    "excitation_energy": "eV",
    "extent": "um",
    "particle_charge": "C",
    "source_beam_width_x": "um",
    "source_beam_width_y": "um",
    "source_polar_angle": "degree",
    "source_azimuth_angle": "degree",
    "source_power": "W",
    "sputter_ion_charge": "C",
    "sputter_source_energy": "eV",
    "sputter_source_beam_current": "A",
    "sputter_source_width_x": "um",
    "sputter_source_width_y": "um",
    "sputter_source_incidence_polar_angle": "degree",
    "sputter_source_azimuth_angle": "degree",
    "analyzer_take_off_polar_angle": "degree",
    "analyzer_take_off_azimuth_angle": "degree",
    "analysis_width_x": "m",
    "analysis_width_y": "m",
    "target_bias": "V",
    "time_correction": "s",
    "work_function": "eV",
    "spatial_acceptance": "um",
    "pass_energy": "eV",
    "differential_width_aes": "eV",
    "dwell_time": "s",
    "sample_normal_polar_angle_of_tilt": "degree ",
    "sample_normal_tilt_azimuth_angle": "degree",
    "sample_rotation_angle": "degree",
    "start_energy": "eV",
    "step_size": "eV",
}


class VamasMapper(XPSMapper):
    """
    Class for restructuring .txt data file from
    Vamas format into python dictionary.
    """

    config_file = "config_vms.json"

    def __init__(self):
        self.multiple_spectra_groups: bool = True
        self.duplicate_spectrum_types: list[str] = []

        super().__init__()

    def _select_parser(self):
        """
        Select parser based on the structure of the Vamas file,
        i.e., whether it is regular or irregular.

        Returns
        -------
        VamasParserVMS
            Vamas parser for reading this file structure.

        """
        return VamasParser()

    def construct_data(self):
        """Map VMS format to NXmpes-ready dict."""

        def get_duplicate_spectrum_types(spectra: list[dict]) -> set:
            """
            Find all duplicate 'spectrum_type' values in the given list of spectra.

            Returns a set of duplicate spectrum types.
            """
            seen = set()
            duplicates = set()

            for spectrum in spectra:
                spectrum_type = spectrum.get("spectrum_type")
                if spectrum_type:
                    if spectrum_type in seen:
                        duplicates.add(spectrum_type)
                    seen.add(spectrum_type)

            return duplicates

        spectra = deepcopy(self.raw_data)

        self._xps_dict["data"]: dict[str, Any] = {}

        if len({spectrum.get("group_name") for spectrum in spectra}) == 1:
            self.multiple_spectra_groups = False

        if not self.multiple_spectra_groups:
            self.duplicate_spectrum_types = get_duplicate_spectrum_types(spectra)

        for spectrum in spectra:
            self._update_xps_dict_with_spectrum(spectrum)

    def _update_xps_dict_with_spectrum(self, spectrum: dict[str, Any]):
        """
        Map one spectrum from raw data to NXmpes-ready dict.
        """
        entry_parts = []

        parts_to_use = ["group_name"] * bool(self.multiple_spectra_groups) + [
            "spectrum_type"
        ]

        for part in parts_to_use:
            val = spectrum.get(part, None)
            if val:
                entry_parts += [val]

        if (
            len(entry_parts) == 1
            and spectrum["spectrum_type"] in self.duplicate_spectrum_types
        ):
            entry_parts += [spectrum["time_stamp"]]

        entry = construct_entry_name(entry_parts)

        if not entry:
            if not self.multiple_spectra_groups:
                entry = f"entry{spectrum['spectrum_id']}"
            else:
                entry = "entry"

        entry_parent = f"/ENTRY[{entry}]"

        for key, value in spectrum.items():
            if key.startswith("entry"):
                entry_parent = "/ENTRY[entry]"
                key = key.replace("entry/", "", 1)
            mpes_key = f"{entry_parent}/{key}"
            if "units" in key:
                value = convert_units(value)
            self._xps_dict[mpes_key] = value
            units = convert_units(get_units_for_key(key, UNITS))
            if units is not None:
                self._xps_dict[f"{mpes_key}/@units"] = units

        # Create key for writing to data.
        scan_key = construct_data_key(spectrum)

        energy = np.array(spectrum["data"]["x"])
        intensity_raw = np.array(spectrum["data"]["y"])
        intensity_cps = np.array(spectrum["data"]["y_cps"])

        if entry not in self._xps_dict["data"]:
            self._xps_dict["data"][entry] = xr.Dataset()

        # Write averaged cycle data to 'data'.
        all_scan_data = [
            np.array(value)
            for key, value in self._xps_dict["data"][entry].items()
            if scan_key.split("_")[0] in key
        ]
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            averaged_scans = np.mean(all_scan_data, axis=0)

        if averaged_scans.size == 1:
            # on first scan in cycle
            averaged_scans = intensity_cps

        try:
            self._xps_dict["data"][entry][scan_key.split("_")[0]] = xr.DataArray(
                data=averaged_scans,
                coords={"energy": energy},
            )
        except ValueError:
            pass

        self._xps_dict["data"][entry][scan_key] = xr.DataArray(
            data=intensity_cps, coords={"energy": energy}
        )

        # Write raw intensities to 'cycle0'.
        self._xps_dict["data"][entry][f"{scan_key}_chan0"] = xr.DataArray(
            data=intensity_raw, coords={"energy": energy}
        )


KEY_MAP = {
    "block_id": "region",
    "sample_id": "sample_name",
    "technique": "analysis_method",
    "source_energy": "excitation_energy",
    "analyzer_mode": "scan_mode",
    "resolution": "pass_energy",
    "transition_label": "transition",
    "abscissa_label": "energy_label",
    "abscissa_units": "energy_units",
    "abscissa_start": "start_energy",
    "abscissa_step": "step_size",
    "variable_label_1": "y_labels_1",
    "variable_units_1": "y_units_1",
    "variable_label_2": "y_labels_2",
    "variable_units_2": "y_units_2",
    "species_label": "element",
}

VALUE_MAP = {
    "analysis_method": convert_measurement_method,
    "energy_label": convert_energy_type,
    "scan_mode": convert_energy_scan_mode,
}


class VamasParser:
    """A parser for reading vamas files."""

    def __init__(self):
        """Construct the vamas parser.

        Class attributes are a VamasHeader, which stores the vamas header
        attributes, blocks, which store the individual Block objects. Each
        block represents one spectrum, then there are several kinds of
        vamas attribute keys, which are used, depending on how the
        vamas file is formatted.
        """
        self.data: list[str] = []

        self.header = VamasHeader()
        self.blocks: list[VamasBlock] = []

    def parse_file(self, file: str | Path, **kwargs):
        """Parse the vamas file into a list of dictionaries.

        Parameters
        ----------
        file: str
           The location and name of the vamas file to be parsed.
        """
        self._read_lines(file)
        self._parse_header()
        self._parse_blocks()
        return self.build_list()

    def _read_lines(self, file: str | Path):
        """Read in vamas text file."""
        with open(file, "rb") as vms_file:
            for line in vms_file:
                if line.endswith((b"\r\n", b"\n")):
                    self.data += [line.decode("utf-8", errors="ignore").strip()]

    def _extract_n_lines_to_list(self, number_of_lines):
        """Ectract n number of lines to a list."""
        extracted = []
        for _ in range(number_of_lines):
            extracted += [self.data.pop(0)]

        return extracted

    def _setattr_with_datacls_type(self, datacls, attr, value):
        """
        Set attribute of dataclass instance with the field_type
        defined in the dataclass.
        """
        field_type = type(getattr(datacls, attr))
        setattr(datacls, attr, field_type(value))

    def _parse_header(self):
        """Parse the Vamas header into a VamasHeader object.

        The common_header_attr are the header attributes that are common
        to all types of Vamas experiment modes.
        Returns
        -------
        None.
        """
        common_attrs = [
            "format_id",
            "institute_id",
            "instrument_model_id",
            "operator_id",
            "experiment_id",
            "num_comment_lines",
        ]
        map_attrs = ["num_analysis_positions", "num_x_coords", "num_y_coords"]

        for attr in common_attrs:
            self._setattr_with_datacls_type(self.header, attr, self.data.pop(0).strip())

        num_comment_lines = int(self.header.num_comment_lines)
        self.header.comment_lines = self._extract_n_lines_to_list(num_comment_lines)

        self.header.exp_mode = check_for_allowed_in_list(
            self.data.pop(0).strip(), EXP_MODES
        )
        self.header.scan_mode = self.data.pop(0).strip()

        if self.header.exp_mode in ["MAP", "MAPDP", "NORM", "SDP"]:
            self.header.num_spectral_regions = int(self.data.pop(0).strip())
        else:
            delattr(self.header, "num_spectral_regions")

        for attr in map_attrs:
            if self.header.exp_mode in ["MAP", "MAPDP"]:
                self._setattr_with_datacls_type(
                    self.header, attr, self.data.pop(0).strip()
                )
            else:
                delattr(self.header, attr)

        self.header.num_exp_var = int(self.data.pop(0).strip())

        for exp_var_no in range(int(self.header.num_exp_var)):
            exp_var = ExpVariable()
            for attr in ["label", "unit"]:
                self._setattr_with_datacls_type(exp_var, attr, self.data.pop(0).strip())
                setattr(
                    self.header, f"exp_var_{exp_var_no}_{attr}", getattr(exp_var, attr)
                )

        self.header.num_entries_in_inclusion_list = int(self.data.pop(0).strip())
        self.header.inclusion_list = self._extract_n_lines_to_list(
            self.header.num_entries_in_inclusion_list
        )

        self.header.num_manually_entered_items_in_block = int(self.data.pop(0).strip())
        self.header.manually_entered_items_in_block = self._extract_n_lines_to_list(
            self.header.num_manually_entered_items_in_block
        )

        self.header.num_future_upgrade_exp_entries = int(self.data.pop(0).strip())
        self.header.num_future_upgrade_block_entries = int(self.data.pop(0).strip())
        self.header.future_upgrade_exp_entries = self._extract_n_lines_to_list(
            self.header.num_future_upgrade_exp_entries
        )

        self.header.num_blocks = int(self.data.pop(0).strip())
        self.header.validate_types()

    def _parse_blocks(self):
        """Parse all (metadata) of Vamas blocks."""
        for _ in range(int(self.header.num_blocks)):
            self.blocks += [self._parse_one_block()]

    def _parse_one_block(self):
        """
        Parse one Vamas Block.

        Returns
        -------
        block : vamas.Block object.
            A block represents one spectrum with its metadata.

        """
        # pylint: disable=too-many-statements
        block = VamasBlock()
        block.block_id = self.data.pop(0).strip()
        block.sample_id = self.data.pop(0).strip()
        block.year = int(self.data.pop(0).strip())
        block.month = int(self.data.pop(0).strip())
        block.day = int(self.data.pop(0).strip())
        block.hour = int(self.data.pop(0).strip())
        block.minute = int(self.data.pop(0).strip())
        block.second = int(self.data.pop(0).strip().split(".")[0])
        block.no_hrs_in_advance_of_gmt = int(self.data.pop(0).strip())
        block.num_comment_lines = int(self.data.pop(0).strip())
        for _ in range(block.num_comment_lines):
            block.comment_lines += [self.data.pop(0)]
        block.technique = check_for_allowed_in_list(
            self.data.pop(0).strip(), ALLOWED_TECHNIQUES
        )

        for attr in ["x_coord", "y_coord"]:
            if self.header.exp_mode in ["MAP", "MAPDP"]:
                self._setattr_with_datacls_type(block, attr, self.data.pop(0).strip())
            else:
                delattr(block, attr)

        for _ in range(int(self.header.num_exp_var)):
            block.exp_var_value = self.data.pop(0).strip()
        block.source_label = self.data.pop(0).strip()

        for attr in [
            "sputter_ion_atomic_number",
            "sputter_ion_num_atoms",
            "sputter_ion_charge",
        ]:
            if self.header.exp_mode in [
                "MAPDP",
                "MAPSVDP",
                "SDP",
                "SDPSV",
            ] or block.technique in [
                "FABMS",
                "FABMS energy spec",
                "ISS",
                "SIMS",
                "SIMS energy spec",
                "SNMS",
                "SNMS energy spec",
            ]:
                self._setattr_with_datacls_type(block, attr, self.data.pop(0).strip())
            else:
                delattr(block, attr)
        block.source_energy = float(self.data.pop(0).strip())
        block.source_power = self.data.pop(0).strip()
        block.source_beam_width_x = self.data.pop(0).strip()
        block.source_beam_width_x = self.data.pop(0).strip()

        for attr in ["field_of_view_x", "field_of_view_y"]:
            if self.header.exp_mode in ["MAP", "MAPDP", "MAPSV", "MAPSVDP", "SEM"]:
                self._setattr_with_datacls_type(block, attr, self.data.pop(0).strip())
            else:
                delattr(block, attr)

        for attr in [
            "first_linescan_x_start",
            "first_linescan_y_start",
            "first_linescan_x_end",
            "first_linescan_y_end",
            "last_linescan_x_end",
            "last_linescan_y_end",
        ]:
            if self.header.exp_mode in ["MAP", "MAPSVDP", "SEM"]:
                self._setattr_with_datacls_type(block, attr, self.data.pop(0).strip())
            else:
                delattr(block, attr)

        block.source_polar_angle = float(self.data.pop(0).strip())
        block.source_azimuth_angle = float(self.data.pop(0).strip())
        block.analyzer_mode = self.data.pop(0).strip()
        block.resolution = float(self.data.pop(0).strip())

        if block.technique == "AES diff":
            self.header.differential_width_aes = float(self.data.pop(0).strip())
        else:
            delattr(block, "differential_width_aes")

        block.magnification = float(self.data.pop(0).strip())
        block.work_function = float(self.data.pop(0).strip())
        block.target_bias = float(self.data.pop(0).strip())
        block.analysis_width_x = float(self.data.pop(0).strip())
        block.analysis_width_y = float(self.data.pop(0).strip())
        block.analyzer_take_off_polar_angle = float(self.data.pop(0).strip())
        block.analyzer_take_off_azimuth_angle = float(self.data.pop(0).strip())
        block.species_label = self.data.pop(0).strip()
        block.transition_label = self.data.pop(0).strip()
        block.particle_charge = int(self.data.pop(0).strip())

        if self.header.scan_mode == "REGULAR":
            block.abscissa_label = self.data.pop(0).strip()
            block.abscissa_units = self.data.pop(0).strip()
            block.abscissa_start = float(self.data.pop(0).strip())
            block.abscissa_step = float(self.data.pop(0).strip())

            block.no_variables = int(self.data.pop(0).strip())
            for var in range(block.no_variables):
                name = "variable_label_" + str(var + 1)
                setattr(block, name, self.data.pop(0).strip())
                name = "variable_units_" + str(var + 1)
                setattr(block, name, self.data.pop(0).strip())

        else:
            block.no_variables = int(self.data.pop(0).strip()) - 1
            block.abscissa_label = self.data.pop(0).strip()
            block.abscissa_units = self.data.pop(0).strip()
            for var in range(block.no_variables):
                name = "variable_label_" + str(var + 1)
                setattr(block, name, self.data.pop(0).strip())
                name = "variable_units_" + str(var + 1)
                setattr(block, name, self.data.pop(0).strip())

        block.signal_mode = self.data.pop(0).strip()
        block.dwell_time = float(self.data.pop(0).strip())
        block.no_scans = int(self.data.pop(0).strip())
        block.time_correction = float(self.data.pop(0).strip())

        for attr in [
            "sputter_source_energy",
            "sputter_source_beam_current",
            "sputter_source_width_x",
            "sputter_source_width_y",
            "sputter_source_incidence_polar_angle",
            "sputter_source_azimuth_angle",
        ]:
            if self.header.exp_mode in [
                "MAPDP",
                "MAPSVDP",
                "SDP",
                "SDPSV",
            ] and block.technique in [
                "AES",
                "AES diff",
                "AES dir",
                "EDX",
                "ELS",
                "UPS",
                "XPS",
                "XRF",
            ]:
                self._setattr_with_datacls_type(block, attr, self.data.pop(0).strip())
            else:
                delattr(block, attr)

        block.sample_normal_polar_angle_of_tilt = float(self.data.pop(0).strip())
        block.sample_normal_tilt_azimuth_angle = float(self.data.pop(0).strip())
        block.sample_rotation_angle = float(self.data.pop(0).strip())
        block.no_additional_params = int(self.data.pop(0).strip())

        for param_no in range(block.no_additional_params):
            param = VamasAdditionalParam()
            for attr in ["label", "unit", "value"]:
                self._setattr_with_datacls_type(param, attr, self.data.pop(0).strip())
                setattr(block, f"param_{param_no}_{attr}", getattr(param, attr))

        block.future_upgrade_block_entries = self._extract_n_lines_to_list(
            self.header.num_future_upgrade_block_entries
        )

        block.num_ord_values = int(self.data.pop(0).strip())
        if self.header.scan_mode == "IRREGULAR":
            del self.data[:2]

        for var_no in range(block.no_variables):
            var = OrdinateValue()
            for attr in ["min_ord_value", "max_ord_value"]:
                self._setattr_with_datacls_type(var, attr, self.data.pop(0).strip())
                setattr(block, f"{attr}_{var_no + 1}", getattr(var, attr))

        self._add_data_values(block)

        block.validate_types()

        return block

    def _add_data_values(self, block: VamasBlock):
        """Add data values to a Vamas data block."""
        if self.header.scan_mode == "REGULAR":
            self._add_regular_data(block)
        elif self.header.scan_mode == "IRREGULAR":
            self._add_irregular_data(block)
        elif self.header.scan_mode == "MAPPING":
            self._add_mapping_data(block)

    def _add_regular_data(self, block: VamasBlock):
        """Parse data with regularly spaced energy axis."""
        data_dict: dict[str, np.ndarray] = {}

        start = float(block.abscissa_start)
        step = float(block.abscissa_step)
        num = int(block.num_ord_values / block.no_variables)
        energy = np.array([round(start + i * step, 2) for i in range(num)])

        if block.abscissa_label == "binding energy":
            energy = np.flip(energy)

        setattr(block, "x", energy)

        for var in range(block.no_variables):
            if var == 0:
                name = "y"
            else:
                name = "y" + str(var)

        data_array = np.array(self.data[: block.num_ord_values], dtype=float)

        self.data = self.data[block.num_ord_values :]

        for var in range(block.no_variables):
            max_var = block.no_variables
            if var == 0:
                name = "y"
            else:
                name = "y" + str(var)
            data_array_slice = data_array[var::max_var]
            data_dict[name] = data_array_slice
            setattr(block, name, data_dict[name])

    def _add_irregular_data(self, block: VamasBlock):
        """Parse data with regularly spaced energy axis."""
        data_dict: dict[str, np.ndarray] = {}

        block_data = np.array(self.data[: block.num_ord_values], dtype=float)

        energy = block_data[:: block.no_variables + 1]
        if block.abscissa_label == "binding energy":
            energy = np.flip(energy)

        setattr(block, "x", energy)
        block.abscissa_start = float(min(energy))
        block.abscissa_step = float(get_minimal_step(energy))

        for var in range(block.no_variables):
            if var == 0:
                name = "y"
            else:
                name = "y" + str(var)

        for var in range(block.no_variables):
            if var == 0:
                name = "y"
            else:
                name = "y" + str(var)
            data_array_slice = block_data[var + 1 :: block.no_variables + 1]
            data_dict[name] = data_array_slice
            setattr(block, name, data_dict[name])

        self.data = self.data[block.num_ord_values :]

    def _add_mapping_data(self, block: VamasBlock):
        """
        Parse data in mapping format.

        TBD!
        """
        pass

    def _get_scan_numbers_for_spectra(self, spectra: list[dict]):
        """
        For a flat list of spectra dictionaries, group the spectra
        by group name and spectrum type and iteratively give them
        scan numbers.

        Parameters
        ----------
        spectra : list
            List of dicts with each dict containing data and metadata
            for one spectrum.

        Returns
        -------
        flattened_spectra : list
            Same list of dicts, but each spectrum gets a scan number.

        """
        grouped_spectra = [
            list(y)
            for x, y in groupby(
                sorted(spectra, key=lambda x: (x["group_name"], x["spectrum_type"])),
                lambda x: (x["group_name"], x["spectrum_type"]),
            )
        ]

        for group in grouped_spectra:
            for i, spectrum in enumerate(group):
                spectrum["scan_no"] = i

        flattened_spectra = [
            spectrum for group in grouped_spectra for spectrum in group
        ]

        return flattened_spectra

    def build_list(self):
        """
        Construct a list of dictionaries from the Vamas objects

        Returns
        -------
        List
            Each list element is a dictionary with the data and
            metadata of one spectrum.

        """
        group_id = -1
        temp_group_name = ""
        spectra = []

        header_dict = {
            convert_pascal_to_snake(k): v for (k, v) in self.header.dict().items()
        }
        del header_dict["comment_lines"]

        header_comments = handle_comments(
            self.header.comment_lines, comment_type="header"
        )

        update_dict_without_overwrite(header_dict, header_comments)

        for spectrum_id, block in enumerate(self.blocks):
            group_name = block.sample_id
            # This set of conditions detects if the group name has changed.
            # If it has, then it increments the group_idx.
            if group_name != temp_group_name:
                temp_group_name = group_name
                group_id += 1

            spectrum_type = str(block.species_label + block.transition_label)
            transitions = [spectrum_type] if spectrum_type else []

            settings = {
                convert_pascal_to_snake(k): v for (k, v) in block.dict().items()
            }

            update_dict_without_overwrite(settings, header_dict)

            settings["n_values"] = int(block.num_ord_values / block.no_variables)

            # Remap to the MPES-prefered keys and values .
            re_map_keys(settings, KEY_MAP)
            re_map_values(settings, VALUE_MAP)

            comment_dict = handle_comments(block.comment_lines, comment_type="block")

            if "casa" in comment_dict:
                casa_process = comment_dict["casa"]

                fit_aux_signals = ["fit_sum"]

                for energy_calibration in casa_process.casa_data["energy_calibrations"]:
                    block.x = energy_calibration.apply_energy_shift(block.x)

                for i, region in enumerate(casa_process.casa_data["regions"]):
                    region.calculate_background(block.x, block.y)
                    region.data_cps = region.data / block.dwell_time
                    fit_aux_signals += [f"background{i}_intensity"]

                for i, component in enumerate(casa_process.casa_data["components"]):
                    component.calculate_lineshape(block.x)
                    component.data_cps = component.data / block.dwell_time
                    fit_aux_signals += [f"peak{i}_intensity"]

                flattened_casa_data = casa_process.flatten_metadata()

                if casa_process.casa_data["components"]:
                    flattened_casa_data["fit_label"] = spectrum_type

                flattened_casa_data["fit_aux_signals"] = fit_aux_signals

                comment_dict.update(flattened_casa_data)
                del comment_dict["casa"]

            re_map_keys(comment_dict, KEY_MAP)
            re_map_values(comment_dict, VALUE_MAP)

            update_dict_without_overwrite(settings, comment_dict)

            # Convert the native time format to the datetime string
            # in the ISO 8601 format
            tzinfo = datetime.timezone(
                datetime.timedelta(hours=block.no_hrs_in_advance_of_gmt)
            )
            try:
                date_time = datetime.datetime(
                    block.year,
                    block.month,
                    block.day,
                    block.hour,
                    block.minute,
                    block.second,
                    tzinfo=tzinfo,
                )
            except ValueError:
                date_time = datetime.datetime(1, 1, 1, 0, 0, 0, tzinfo=tzinfo)

            if isinstance(date_time, datetime.datetime):
                date_time = date_time.isoformat()

            # Map x-y values to 2D lists.
            settings["extent"] = np.array(
                [
                    settings["source_beam_width_x"],
                    settings["source_beam_width_y"],
                ],
                dtype=float,
            )
            settings["spatial_acceptance"] = np.array(
                [
                    settings["analysis_width_x"],
                    settings["analysis_width_y"],
                ],
                dtype=float,
            )

            data = {"x": block.x}

            for var in range(int(block.no_variables)):
                if var == 0:
                    key = "y"

                    data["y"] = getattr(block, "y")
                    del settings["y"]

                    if block.variable_label_1 in ["Intensity", "counts", "count rate"]:
                        y_cps = [np.round(y / block.dwell_time, 2) for y in block.y]
                        data["y_cps"] = np.array(y_cps)

                else:
                    key = "y" + str(var)
                    data[key] = getattr(block, key)
                    del settings[key]

            spec_dict = {
                "time_stamp": date_time,
                "group_name": group_name,
                "group_id": group_id,
                "spectrum_type": spectrum_type,
                "transitions": transitions,
                "spectrum_id": spectrum_id,
                "scans": block.no_scans,
                "data": data,
            }

            remove_keys = [
                "comment_lines",
                "year",
                "month",
                "day",
                "hour",
                "minute",
                "second",
                "no_hrs_in_advance_of_gmt",
                "source_beam_width_x",
                "source_beam_width_y",
                "x",
            ]

            drop_unused_keys(settings, remove_keys)

            spec_dict.update(settings)
            spectra += [spec_dict]

        spectra = self._get_scan_numbers_for_spectra(spectra)

        return spectra



# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Backgrounds for peak fitting.
"""

from collections.abc import Callable
from typing import Optional, cast

import numpy as np

# Workaround for different functions for trapezoid in numpy
import numpy.typing as npt
import scipy

NPTrapezoidType = Callable[..., np.typing.NDArray[np.float64]]

if np.lib.NumpyVersion(np.__version__) >= "2.0.0":
    np_trapezoid = cast(NPTrapezoidType, getattr(np, "trapezoid"))
else:
    np_trapezoid = cast(NPTrapezoidType, np.trapz)  # noqa: NPY201


class LinearBackground:
    """Linear background model for XPS spectra that connects the first and last points in the y array."""

    def __init__(self, *args):
        """
        Initialize the linear background model without predefined parameters.
        The slope and intercept will be calculated based on the input data.
        """
        self.slope = None
        self.intercept = None

    def calc_background(self, x: np.ndarray, y: np.ndarray) -> np.ndarray:
        """
        Calculate the linear background based on the first and last points of the y array.

        Parameters:
            x (np.ndarray): The energy values at which to evaluate the background.
            y (np.ndarray): The observed intensity values at the corresponding energy points in x.

        Returns:
            np.ndarray: The background intensity at each energy point.
        """
        if len(x) != len(y):
            raise ValueError("x and y arrays must have the same length.")

        # Calculate slope and intercept using the first and last points
        self.slope = (y[-1] - y[0]) / (x[-1] - x[0])
        self.intercept = y[0] - self.slope * x[0]

        # Return the background values using the linear equation
        return self.slope * x + self.intercept

    def formula(self) -> str:
        """
        Returns the formula used for the linear background model.

        Returns:
            str: The formula for the linear background.
        """
        return (
            "Linear Background Formula:\n"
            "B(x) = slope * x + intercept\n"
            "Where:\n"
            "  slope: (y[-1] - y[0]) / (x[-1] - x[0])  # Calculated from the first and last points of y\n"
            "  intercept: y[0] - slope * x[0]           # Calculated from the first point of y\n"
            "  x: Energy values\n"
            "  y: Intensity values\n"
        )


class StepUp:
    """Step Up background model for XPS spectra, based on the complementary error function."""

    def __init__(self, a0: float, a1: float, a2: float, a3: float, *args):
        """
        Initialize the Step Up background model without predefined parameters.
        """
        self.a0 = a0  # Step magnitude
        self.a1 = a1  # Edge position
        self.a2 = a2  # FWHM-related parameter
        self.a3 = a3  # Constant offset

    def calc_background(self, x: np.ndarray, y: np.ndarray) -> np.ndarray:
        """
        Calculate the Step Up background based on the complementary error function.

        Parameters:
            x (np.ndarray): The energy values at which to evaluate the background.
            a0 (float): Magnitude of the step.
            a1 (float): Location of the step (edge position).
            a2 (float): Parameter related to FWHM (2 * sqrt(ln(2)) * a2 = FWHM).
            a3 (float): Constant offset.

        Returns:
            np.ndarray: The background intensity at each energy point.
        """
        return (self.a0 / 2) * scipy.special.erfc((self.a1 - x) / self.a2) + self.a3

    def formula(self) -> str:
        """
        Returns the formula used for the Step Up background model.

        Returns:
            str: The formula for the Step Up background.
        """
        return (
            "Step Up Background Formula:\n"
            "B(x; a0, a1, a2, a3) = (a0 / 2) * erfc((a1 - x) / a2) + a3\n"
            "Where:\n"
            "  a0: Magnitude of the step\n"
            "  a1: Edge position\n"
            "  a2: FWHM-related parameter (FWHM = 2 * sqrt(ln(2)) * a2)\n"
            "  a3: Constant offset\n"
        )


class StepDown:
    """Step Down background model for XPS spectra, the reflection of Step Up around the edge position."""

    def __init__(self, a0: float, a1: float, a2: float, a3: float, *args):
        """
        Initialize the Step Up background model without predefined parameters.
        """
        self.a0 = a0  # Step magnitude
        self.a1 = a1  # Edge position
        self.a2 = a2  # FWHM-related parameter
        self.a3 = a3  # Constant offset

    def calc_background(self, x: np.ndarray, y: np.ndarray) -> np.ndarray:
        """
        Calculate the Step Down background as the reflection of the Step Up background.

        Parameters:
            x (np.ndarray): 1D array of energy values at which to evaluate the background.
            a0 (float): Magnitude of the step.
            a1 (float): Location of the step (edge position).
            a2 (float): Parameter related to FWHM (2 * sqrt(ln(2)) * a2 = FWHM).
            a3 (float): Constant offset.

        Returns:
            np.ndarray: 1D array of background intensity at each energy point.
        """

        return (self.a0 / 2) * scipy.special.erfc((x - self.a1) / self.a2) + self.a3

    def formula(self) -> str:
        """
        Returns the formula used for the Step Down background model.

        Returns:
            str: The formula for the Step Down background.
        """
        return (
            "Step Down Background Formula:\n"
            "B(x) = (a0 / 2) * erfc((x - a1) / a2) + a3\n"
            "Where:\n"
            "  a0: Magnitude of the step\n"
            "  a1: Edge position\n"
            "  a2: FWHM-related parameter (FWHM = 2 * sqrt(ln(2)) * a2)\n"
            "  a3: Constant offset\n"
        )


class Shirley:
    """
    Shirley background subtraction using the Sherwood method.

    The Shirley method is used to compute the background of an x-ray or
    spectroscopy data set by iteratively subtracting the baseline until
    convergence. This class implements this method with options for
    tolerance and maximum iterations.
    """

    def __init__(self, *args) -> None:
        """Initialize the Shirley background subtraction object."""
        pass

    def calc_background(
        self, x: np.ndarray, y: np.ndarray, tol: float = 1e-5, maxit: int = 15
    ) -> np.ndarray:
        """
        Calculate the Shirley background using the Sherwood method.

        Parameters:
        - x (np.ndarray): The x-axis data (energy or time, etc.).
        - y (np.ndarray): The y-axis data (intensity or counts, etc.).
        - tol (float, optional): The tolerance for the convergence criterion. Defaults to 1e-5.
        - maxit (int, optional): The maximum number of iterations to prevent infinite loops. Defaults to 15.

        Raises:
        - ValueError: If input arrays have mismatched dimensions, incorrect types, or the fit does not converge.

        Returns:
        - np.ndarray: The calculated Shirley background.
        """
        # Validate input
        if not isinstance(x, np.ndarray) or not isinstance(y, np.ndarray):
            raise ValueError(
                f"Parameters x and y must be numpy arrays, not {type(x)} and {type(y)}"
            )

        if len(x) != len(y):
            raise ValueError("x and y arrays must have the same length.")

        if x.size == 0 or y.size == 0:
            raise ValueError("x or y array is empty.")

        if x.ndim > 1 or y.ndim > 1:
            raise ValueError(
                f"Data arrays must be one-dimensional. Found shapes x: {x.shape}, y: {y.shape}."
            )

        # Reverse the data if it is in ascending order
        is_reversed = False
        if x[0] < x[-1]:
            is_reversed = True
            x = np.flip(x)
            y = np.flip(y)

        # Initialize background arrays
        background = np.zeros_like(x)
        background_next = np.zeros_like(x)

        # Iterative loop to compute Shirley background
        for iters in range(maxit):
            k = (y[0] - y[-1]) / np_trapezoid(y - background, x=x)

            # Update the background using trapezoidal integration
            for energy in range(len(x)):
                background_next[energy] = k * np_trapezoid(
                    y[energy:] - background[energy:], x=x[energy:]
                )

            # Check for convergence
            diff = np.linalg.norm(background_next - background)
            background = np.copy(background_next)
            if diff < tol:
                break
        else:
            # Raise an error if the maximum iterations are reached
            raise ValueError(
                "Maximum number of iterations exceeded before convergence."
            )

        # Reverse back the result if the data was originally reversed
        return np.flip(y[-1] + background) if is_reversed else y[-1] + background

    def formula(self) -> str:
        """
        Returns the iterative formula used in the Shirley background subtraction.

        Returns:
        - str: The formula used for the Shirley background subtraction.
        """
        return (
            "Shirley Background Subtraction Formula:\n"
            "1. Initial background B_i = 0 for all i.\n"
            "2. In each iteration, calculate:\n"
            "   k = (y_0 - y_n) / integral(x_0, x_n, (y_j - B_j) dx_j)\n"
            "3. Update the background using the formula:\n"
            "   B_i = k * integral(x_i, x_n, (y_j - B_j) dx_j)\n"
            "4. Repeat steps 2 and 3 until convergence (|B_next - B| < tol).\n"
            "5. If convergence fails after maxit iterations, raise an error.\n"
            "Where:\n"
            "  B_i: Background at point i\n"
            "  k: Scaling factor based on the integral\n"
            "  y_i: Original data at point i\n"
            "  x_i: X-axis values\n"
            "  B: Current background estimate\n"
            "  tol: Convergence tolerance\n"
            "  maxit: Maximum number of iterations"
        )


class TougaardU3:
    """U3 Tougaard background model for XPS spectra.

    Parameters:
        E0 (float): Energy onset for the background, typically near the core-level peak energy.
        A (float): Scaling factor for the background intensity.
        C (float): Shape parameter that determines the background curvature.
    """

    def __init__(self, E0: float, A: float, C: float, *args):
        """
        Initialize the U3 Tougaard model with the required parameters.

        Parameters:
            E0 (float): Energy onset for the background, typically near the core-level peak energy.
            A (float): Scaling factor for the background intensity.
            C (float): Shape parameter that determines the background curvature.
        """
        self.E0 = E0
        self.A = A
        self.C = C

    def calc_background(self, x: np.ndarray, y: np.ndarray) -> np.ndarray:
        """Calculate the Tougaard background at each energy point.

        Args:
            x (np.ndarray): Array of energy values.

        Returns:
            np.ndarray: Tougaard background values corresponding to each energy value in x.
        """
        return self.A / ((x - self.E0) ** 2 + self.C)

    def formula(self) -> str:
        """
        Returns the formula used for the Tougaard background model.

        Returns:
            str: The formula for the Tougaard background.
        """
        return (
            "Tougaard U3 Background Formula:\n"
            "B(x; A, E0, C) = A / ((x - E0)^2 + C)\n"
            "Where:\n"
            "  B(x): Background at energy x\n"
            "  A: Scaling factor\n"
            "  E0: Energy onset (near core-level peak energy)\n"
            "  C: Shape parameter (determines background curvature)\n"
        )

    def __repr__(self) -> str:
        """
        Returns a string representation of the GaussianLorentzianSum object, including details
        for position, width, intensity, and fraction_gauss.

        Returns:
        - str: The string representation of the object.
        """
        return f"TougaardU3(E0={self.E0}, A={self.A}, C={self.C})"


class TougaardU4:
    """U4 Tougaard background model for XPS spectra."""

    def __init__(
        self, B: float, C: float, D: float, Eg: float, temp: float = 300.0, *args
    ):
        """
        Initialize the U4 Tougaard model with the required parameters.

        Parameters:
            B (float): A scaling factor that adjusts the overall amplitude of the background.
            C (float): A parameter influencing the background's shape, specifically the width of the energy region where the background decays.
            D (float): Another shape parameter that controls the fall-off of the background, influencing the tail of the background.
            Eg (float): The energy onset for the background, typically near the core-level peak energy, controlling where the background starts.
            temp (float): Temperature in Kelvin (default is 300 K).
        """
        self.B = B
        self.C = C
        self.D = D
        self.Eg = Eg
        self.temp = temp

    def calc_background(self, x: np.ndarray, y: np.ndarray) -> np.ndarray:
        """
        Calculate the modified Tougaard background at each energy point using a modified expression.

        Parameters:
            x (np.ndarray or float): The energy values at which to evaluate the background.

        Returns:
            np.ndarray or float: The background intensity at each energy point.
        """
        kb = 0.000086  # Boltzmann constant in eV/K

        return (
            (self.B * x)
            / ((self.C - x**2) ** 2 + self.D * x**2)
            * 1
            / (np.exp((self.Eg - x) / (self.temp * kb)) + 1)
        )

    def formula(self) -> str:
        """
        Returns the formula used for the U4 Tougaard background model.

        Returns:
            str: The formula for the U4 Tougaard background.
        """
        return (
            "U4 Tougaard Background Formula:\n"
            "B(x; B, C, D, E_g, T) = (B * x) / ((C - x^2)^2 + D * x^2) * 1 / (exp((Eg - x) / (kB * T)) + 1)\n"
            "Where:\n"
            "  B: Scaling factor\n"
            "  C: Shape parameter (affects width of energy decay region)\n"
            "  D: Shape parameter (controls the tail of the background)\n"
            "  Eg: Energy onset (near core-level peak energy)\n"
            "  kB: Boltzmann constant (0.000086 eV/K)\n"
            "  T: Temperature (default 300 K, can be modified)\n"
            "  x: Energy values\n"
        )

    def __repr__(self) -> str:
        """
        Returns a string representation of the GaussianLorentzianSum object, including details
        for position, width, intensity, and fraction_gauss.

        Returns:
        - str: The string representation of the object.
        """
        return (
            f"TougaardU4(B={self.B}, C={self.C}, D={self.D}, "
            f"Eg={self.Eg}, temp={self.temp})"
        )



# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# pylint: disable=too-many-lines,too-few-public-methods
"""
Utility function for mapping keys and values in the pynxtools template.
"""

import datetime
import re
from collections.abc import Callable
from typing import Any

ENERGY_TYPE_MAP: dict[str, str] = {
    "BE": "binding",
    "KE": "kinetic",
    "Binding": "binding",
    "Binding Energy": "binding",
    "binding energy": "binding",
    "binding_energy": "binding",
    "Kinetic": "kinetic",
    "Kinetic Energy": "kinetic",
    "kinetic energy": "kinetic",
    "kinetic_energy": "kinetic",
    "Analyser Energy": "binding",
    "analyser_energy": "binding",
}

ENERGY_SCAN_MODE_MAP: dict[str, str] = {
    "Fixed": "fixed_energy",
    "fixed": "fixed_energy",
    "FixedEnergies": "fixed_energy",
    "Swept": "fixed_analyzer_transmission",
    "FixedAnalyzerTransmission": "fixed_analyzer_transmission",
    "FAT": "fixed_analyzer_transmission",
    "FixedRetardationRatio": "fixed_retardation_ratio",
    "FRR": "fixed_retardation_ratio",
    "Snapshot": "snapshot",
    "SnapshotFAT": "snapshot",
}

MEASUREMENT_METHOD_MAP: dict[str, str] = {
    "XPS": "X-ray photoelectron spectroscopy (XPS)",
    "UPS": "ultraviolet photoelectron spectroscopy (UPS)",
    "ElectronSpectroscopy": "electron spectroscopy for chemical analysis (ESCA)",
    "NAPXPS": "near ambient pressure X-ray photoelectron spectroscopy (NAPXPS)",
    "ARXPS": "angle-resolved X-ray photoelectron spectroscopy (ARXPS)",
}

ACQUSITION_MODE_MAP: dict[str, str] = {
    "Image": "pulse counting",
    "Events": "pulse counting",
}

SLIT_TYPE_MAP: dict[str, str] = {"Straight": "straight slit", "Curved": "curved slit"}

BOOL_MAP: dict[str, bool] = {
    "yes": True,
    "Yes": True,
    "no": False,
    "No": False,
    "On": True,
    "Off": False,
}

UNIT_MAP: dict[str, str] = {
    "a.u.": "counts",
    "Counts": "counts",
    "counts/s": "counts_per_second",
    "CPS": "counts_per_second",
    "u": "um",
    "KV": "kV",
    "seconds": "s",
    "(min)": "min",
    "Percent": "",  # should be changed back to percent once pint is updated
    "atom": "dimensionless",
    "eV/atom": "eV",
    "microm": "micrometer",
    "d": "degree",
    "nU": "V",  # workaround for SPECS SLE reader
    "s-1": "1/s",  # workaround for SPECS XY reader
    "norm": None,  # workaround for SPECS XY reader
}


def _replace_from_map(value: Any, value_map: dict[str, Any]):
    """
    For a given value, return a new value if the value is
    part of the value_map.
    """
    return value_map.get(value, value)


def make_converter(value_map: dict[str, Any]) -> Callable[[Any], Any]:
    return lambda value: _replace_from_map(value, value_map)


convert_energy_type = make_converter(ENERGY_TYPE_MAP)
convert_energy_scan_mode = make_converter(ENERGY_SCAN_MODE_MAP)
convert_measurement_method = make_converter(MEASUREMENT_METHOD_MAP)
convert_detector_acquisition_mode = make_converter(ACQUSITION_MODE_MAP)
convert_slit_type = make_converter(SLIT_TYPE_MAP)
convert_bool = make_converter(BOOL_MAP)
convert_units = make_converter(UNIT_MAP)


def get_units_for_key(unit_key: str, unit_map: dict[str, str]) -> str:
    """
    Get correct units for a given key from a dictionary with unit map.
    Parameters
    ----------
    unit_key : str
       Key of type <mapping>:<spectrum_key>, e.g.
       detector/detector_voltage
    Returns
    -------
    str
        Unit for that unit_key.
    """
    regex_match = re.search(r"\[([A-Za-z0-9_]+)\]", unit_key)
    if regex_match is None:
        return unit_map.get(unit_key, None)
    return regex_match.group(1)


def parse_datetime(
    datetime_string: str,
    possible_date_formats: list[str],
    tzinfo: datetime.tzinfo = datetime.timezone.utc,
) -> str:
    """
    Convert a date string to ISO 8601 format with optional timezone handling.

    Convert the native time format to the datetime string
    in the ISO 8601 format: '%Y-%b-%dT%H:%M:%S.%fZ'.
    For different vendors, there are different possible date formats,
    all of which can be checked with this method.
    Optionally, a timezone (tzinfo) can be applied to the datetime object if provided.

    Parameters
    ----------
    datetime_string : str
        String representation of the date
    possible_date_formats : List[str]
        List of possible date time formats to attempt for parsing.
    tzinfo: datetime.tzinfo
        A tzinfo object specifying the desired timezone to apply to the datetime object.
        Defaults to UTC (datetime.timezone.utc).

    Raises
    ------
    ValueError
        If the time format cannot be converted, a ValueError is raised.

    Returns
    -------
    str
        Datetime in ISO 8601 format.
    """
    for date_fmt in possible_date_formats:
        if date_fmt == "%Y-%m-%dT%H:%M:%S.%f%z":
            # strptime only supports six digits for microseconds
            datetime_string = datetime_string[:-7] + datetime_string[-6:]

        try:
            datetime_obj = datetime.datetime.strptime(datetime_string, date_fmt)

            if tzinfo is not None:
                # Apply the specified timezone to the datetime object
                datetime_obj = datetime_obj.replace(tzinfo=tzinfo)

            # Convert to ISO 8601 format
            return datetime_obj.isoformat()

        except ValueError:
            continue

    raise ValueError(
        f"Datetime {datetime_string} could not be converted to ISO 8601 format, "
        f"as it does not match any of these formats: {possible_date_formats}."
    )



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
MKdocs macros for the documentation
"""


def define_env(env):
    """
    This is the hook for defining variables, macros and filters

    - variables: the dictionary that contains the environment variables
    - macro: a decorator function, to declare a macro.
    - filter: a function with one of more arguments,
        used to perform a transformation
    """

    # add to the dictionary of variables available to markdown pages:
    env.variables["version"] = "2023.10"  # Figure out from setuptools-scm eventually



# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# pylint: disable=too-many-instance-attributes
"""
Comment handler for .vms files.
"""

from typing import Any, Literal, Union

from pynxtools_xps.kratos.metadata_kratos import KratosParser
from pynxtools_xps.phi.spe_pro_phi import PhiParser
from pynxtools_xps.reader_utils import convert_pascal_to_snake, split_value_and_unit
from pynxtools_xps.vms.casa_data_model import CasaProcess


def handle_comments(
    comment_list: list[str], comment_type: Literal["header", "block"]
) -> dict[str, Any]:
    comments = {}

    special_keys: tuple[str, ...] = (
        "Casa Info Follows",
        "SpecsLab Prodigy",
        "SOFH",
        "Creation",
    )

    for keyword in special_keys:
        # if any(keyword in line for line in comment_list):
        indices = [i for i, line in enumerate(comment_list) if keyword in line]

        if indices:
            index = indices[0]
            if keyword == "Casa Info Follows":
                if comment_type == "header":
                    special_comments = [comment_list[index]]
                if comment_type == "block":
                    special_comments = comment_list[index:]
                    # TODO: only extract casa comments
                casa_comments, no_of_casa_lines = _handle_casa_comments(
                    special_comments, comment_type
                )
                comments.update(casa_comments)
                del comment_list[index:no_of_casa_lines]

            if keyword == "SpecsLab Prodigy":
                special_comments = [comment_list[index]]
                comment_list = comment_list[index + 1 :]
                comments.update(_handle_prodigy_header_comments(special_comments))

            if keyword == "SOFH":
                end_index = [
                    i for i, line in enumerate(comment_list) if "EOFH" in line
                ][0]
                special_comments = comment_list[index : end_index + 1]  # type: ignore[assignment]
                del comment_list[index : end_index + 1]
                comments.update(_handle_phi_header_comments(special_comments))

            if keyword == "Creation":
                end_index = [
                    i for i, line in enumerate(comment_list) if "X-ray Power" in line
                ][0]
                special_comments = comment_list[index : end_index + 1]  # type: ignore[assignment]
                del comment_list[index : end_index + 1]
                comments.update(_handle_kratos_block_comments(special_comments))

    # Handle non-special comments.
    comments.update(_handle_misc_comments(comment_list))

    return comments


def _handle_casa_comments(
    comment_list: list[str], comment_type: Literal["header", "block"]
) -> tuple[dict[str, Any], int]:
    """Handle comments from CasaXPS, depending on their location."""
    handle_functions = {
        "header": _handle_casa_header_comments,
        "block": _handle_casa_block_comments,
    }

    return handle_functions[comment_type](comment_list)


def _handle_casa_header_comments(comment_list: list[str]) -> tuple[dict[str, Any], int]:
    """Get information about CasaXPS version."""
    comment_line = comment_list[0]

    comments = {
        "casa_version": comment_line.split("Casa Info Follows CasaXPS Version")[
            1
        ].strip()
    }
    no_of_casa_lines = 1

    return comments, no_of_casa_lines


def _handle_casa_block_comments(comment_list: list[str]) -> tuple[dict[str, Any], int]:
    """Get all processing and fitting data from CasaXPS comments."""
    casa = CasaProcess()
    casa.process_comments(comment_list)

    comments = {"casa": casa}

    return comments, casa.no_of_casa_lines


def _handle_prodigy_header_comments(comment_list: list[str]) -> dict[str, str]:
    """Get information about SpecsLab Prodigy version."""
    comment_line = comment_list[0]
    return {"prodigy_version": comment_line.split("Version")[1].strip()}


def _handle_phi_header_comments(comment_list: list[str]) -> dict[str, Any]:
    """Get metadata from Phi system."""
    phi_parser = PhiParser()
    phi_parser.parse_header_into_metadata(comment_list)

    phi_comments = phi_parser.metadata.dict()

    regions = phi_parser.parse_spectral_regions(comment_list)
    areas = phi_parser.parse_spatial_areas(comment_list)

    for region in regions:
        for area in areas:
            concatenated = {**region.dict(), **area.dict()}

        phi_comments.update(concatenated)

    return phi_comments


def _handle_kratos_block_comments(comment_list: list[str]) -> dict[str, Any]:
    """Get metadata from Kratos system."""
    kratos_parser = KratosParser()
    kratos_parser.parse_header_into_metadata(comment_list)

    return kratos_parser.flatten_metadata()


def _handle_misc_comments(
    comment_list: list[str],
) -> dict[str, int | float | str | Any]:
    """Handle any other comments."""
    comments = {}
    for sep in ("=", ":"):
        for line in comment_list:
            try:
                key, value_str = [part.strip(" ") for part in line.split(sep, 1)]
                key = convert_pascal_to_snake(key)
                value, unit = split_value_and_unit(value_str)

                comments[key] = value
                if unit:
                    comments[f"{key}/@units"] = unit
            except ValueError:
                continue
    return comments



# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Helper functions for populating NXmpes template
"""

import logging
import os
import re
from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Optional, Union

import numpy as np
import pint
from pynxtools.units import ureg
from scipy.interpolate import interp1d

logger = logging.getLogger(__name__)


@dataclass
class XpsDataclass:
    """Generic class to hold a data model and a type validation method."""

    def validate_types(self):
        ret = True
        for field_name, field_def in self.__dataclass_fields__.items():
            actual_type = type(getattr(self, field_name))
            if actual_type != field_def.type:
                logger.warning(
                    f"Type mismatch in dataclass {type(self).__name__}. {field_name}: '{actual_type}' instead of '{field_def.type}'"
                )
                ret = False
        return ret

    def __post_init__(self):
        if not self.validate_types():
            raise ValueError(f"Type mismatch in dataclass {type(self).__name__}")

    def dict(self):
        return self.__dict__.copy()


class XPSMapper(ABC):
    """Abstract base class from mapping from a parser to NXmpes template"""

    def __init__(self):
        self.file: str | Path = ""
        self.raw_data: list[str] = []
        self._xps_dict: dict[str, Any] = {}

        self.parser = None

    @abstractmethod
    def _select_parser(self):
        """
        Select the correct parser for the file extension and format.

        Should be implemented by the inheriting mapper.
        """

    @property
    def data_dict(self) -> dict:
        """Getter property."""
        return self._xps_dict

    def parse_file(self, file: str | Path, **kwargs):
        """
        Parse the file using the Scienta TXT parser.

        """
        self.file = file
        self.parser = self._select_parser()
        self.raw_data = self.parser.parse_file(file, **kwargs)

        self._xps_dict["File"] = file
        self._xps_dict["file_ext"] = os.path.splitext(file)[1]

        self.construct_data()

        return self.data_dict

    @abstractmethod
    def construct_data(self):
        """
        Map from individual parser format to NXmpes-ready dict.

        Should be implemented by the inheriting mapper.

        """


def convert_snake_to_pascal(str_value: str):
    """Convert snakecase text to pascal case."""
    return str_value.replace("_", " ").title().replace(" ", "")


def convert_pascal_to_snake(str_value: str):
    """Convert pascal case text to snake case."""
    # Convert CamelCase to snake_case
    snake_case = re.sub(r"(?<!^)(?=[A-Z])", "_", str_value)

    # Convert whitespace to underscores and remove extra underscores
    snake_case_cleaned = re.sub(r"\s+", "_", snake_case).replace("__", "_")

    return snake_case_cleaned.lower()


def safe_arange_with_edges(start: float, stop: float, step: float) -> np.ndarray:
    """
    In order to avoid float point errors in the division by step.

    Args:
        start (float): Lower limit.
        stop (float): Upper limit.
        step (float): Step size between points.
    Returns:
        ndarray
            1D array with values in the interval (start, stop),
            incremented by step.

    """
    return step * np.arange(start / step, (stop + step) / step)


def check_uniform_step_width(lst: list[float]) -> bool:
    """
    Check to see if a non-uniform step width is used in an list.

    Args:
        lst (list): List of data points.

    Returns:
        bool: False if list is non-uniformally spaced.

    """
    start = lst[0]
    stop = lst[-1]
    step = get_minimal_step(lst)

    if step != 0.0 and np.abs((stop - start) / step) > len(lst):
        return False
    return True


def get_minimal_step(lst: list[float] | np.ndarray) -> float:
    """
    Return the minimal difference between two consecutive values
    in a list. Used for extracting minimal difference in a
    list with non-uniform spacing.

    Args:
        lst (list): List of data points.

    Returns:
        step (float): Non-zero, minimal distance between consecutive data
        points in lst.

    """
    lst1 = np.roll(lst, -1)
    diff = np.abs(np.subtract(lst, lst1))
    step = round(np.min(diff[diff != 0]), 2)

    return step


def _resample_array(y: np.ndarray, x0: np.ndarray, x1: np.ndarray) -> np.ndarray:
    """
    Resample an array (y) which has the same initial spacing
    of another array(x0), based on the spacing of a new
    array(x1).

    Args:
        y (np.ndarray): Lineshape array or list.
        x0 (np.ndarray): x array with old spacing.
        x1 (np.ndarray): x array with new spacing.

    Returns:
        np.ndarray: Interpolated y array.

    """
    # pylint: disable=invalid-name
    interp_fn = interp1d(x0, y, axis=0, fill_value="extrapolate")
    return interp_fn(x1)


def interpolate_arrays(x: list[float], array_list: list[np.ndarray]):
    """
    Interpolate data points in case a non-uniform step width was used.

    Parameters
    ----------
    x : list
        List of non-uniformally spaced data points.
    array_list : list
        List of arrays to be interpolated according to new x axis.

    Returns
    -------
    x, array_list
        Interpolated x axis and list of arrays

    """
    # pylint: disable=invalid-name
    if not isinstance(array_list, list):
        array_list = [array_list]
    start = x[0]
    stop = x[-1]
    step = get_minimal_step(x)
    if start > stop:
        # pylint: disable=arguments-out-of-order
        new_x = np.flip(safe_arange_with_edges(stop, start, step))
    else:
        new_x = safe_arange_with_edges(start, stop, step)

    output_list = [_resample_array(arr, np.array(x), new_x) for arr in array_list]

    return new_x, output_list


def check_for_allowed_in_list(value: Any, allowed_values: list[Any]):
    """
    Check if a value is a list of values.
    If not, raise Exception.
    """

    if value not in allowed_values:
        raise Exception(f"{value} not in allowed values: {allowed_values}.")
    return value


def re_map_keys(dictionary: dict[str, Any], key_map: dict[str, str]) -> dict[str, Any]:
    """
    Map the keys in a dictionary such that they are replaced by the values
    in key_map and return the dictionary with replaced keys.

    This is often used to map some metadata keys in a vendor file format
    to the common metadata names. used in NXmpes/NXxps.

    Args:
        dictionary (dict[str, Any]): Dictionary with XPS metadata.
        key_map (dict[str, str]): Mapping of keys from vendor file format
        to common metadata names.

        Example from the VMS parser:
            key_map = {
               "block_id": "region",
               "sample_id": "sample_name",
               "technique": "analysis_method",
               "source_energy": "excitation_energy",
            }

    Returns:
        dictionary (dict[str, Any]): Dictionary with changed keys.

    """
    for key in key_map:
        if key in dictionary:
            map_key = key_map.get(key, key)
            dictionary[map_key] = dictionary.pop(key)
    return dictionary


def re_map_values(
    dictionary: dict[str, Any], map_functions: dict[str, Any]
) -> dict[str, Any]:
    """
    Map the values in a dicitionary using functions defined in a
    mapping functions dictionary.

    This is often used to map some metadata in a vendor file format
    to the enumerations used in NXmpes/NXxps.

    Args:
        dictionary (dict[str, Any]): Dictionary with XPS metadata.
        map_functions (dict[str, Any]): Mapping functions for keys in the dictionary.

        Example from the SLE parser:
        map_functions = {
            "energy/@type": self._change_energy_type,
            "excitation_energy": self._convert_excitation_energy,
            "time_stamp": self._convert_date_time,
            "energy_scan_mode": self._convert_energy_scan_mode,
        }

    Returns:
        dictionary (dict[str, Any]): Dictionary with changed values.

    """
    for key, map_fn in map_functions.items():
        if key in dictionary:
            dictionary[key] = map_fn(dictionary[key])
    return dictionary


def _re_map_single_value(
    input_key: str,
    value: str | int | float | bool | np.ndarray | None,
    map_functions: dict[str, Any],
    **kwargs,
) -> str | int | float | bool | np.ndarray | None:
    """
    Map the values returned from the file to the preferred format for
    the parser output.

    """
    if isinstance(value, str) and value is not None:
        value = value.rstrip("\n")

    for key, map_method in map_functions.items():
        if key == input_key:
            map_method = map_functions[key]
            return map_method(value, **kwargs) if kwargs else map_method(value)  # type: ignore[operator]

    return value


def _check_valid_value(value: str | int | float | bool | np.ndarray) -> bool:
    """
    Check if a value is valid.

    Strings and arrays are considered valid if they are non-empty.
    Numbers and booleans are always considered valid.

    Args:
        value (str | int | float | bool | np.ndarray):
            The value to check. Can be a scalar, string, or NumPy array.

    Returns:
        bool: True if the value is valid, False otherwise.
    """
    if isinstance(value, str | int | float) and value is not None:
        return True
    if isinstance(value, bool):
        return True
    if isinstance(value, np.ndarray) and value.size != 0:
        return True
    return False


def drop_unused_keys(dictionary: dict[str, Any], keys_to_drop: list[str]) -> None:
    """
    Remove unwanted keys from a dictionary.

    Args:
        dictionary (dict[str, Any]):
            Dictionary containing data and metadata for a spectrum.
        keys_to_drop (list[str]):
            List of keys that should be removed from the dictionary.

    Returns:
        None
    """
    for key in keys_to_drop:
        dictionary.pop(key, None)


def update_dict_without_overwrite(d1: dict[str, Any], d2: dict[str, Any]):
    """Update d1 with d2, but don't overwrite existing keys."""
    d1.update({k: v for k, v in d2.items() if k not in d1})


def construct_data_key(spectrum: dict[str, Any]) -> str:
    """
    Construct a key for the 'data' field of the xps_dict.
    Output example: cycle0_scan0.

    """
    if "loop_no" in spectrum:
        cycle_key = f"cycle{spectrum['loop_no']}"
    else:
        cycle_key = "cycle0"

    if "scan_no" in spectrum:
        scan_key = f"scan{spectrum['scan_no']}"
    else:
        scan_key = "scan0"

    return f"{cycle_key}_{scan_key}"


def align_name_part(name_part: str):
    """Make one part of the entry name compliant with NeXus standards."""
    translation_table = str.maketrans(
        {
            " ": "_",
            ",": "",
            ".": "_",
            "-": "_",
            ":": "_",
            "+": "_",
            "/": "_",
            "=": "",
        }
    )

    return name_part.translate(translation_table)


def construct_entry_name(parts: list[str]) -> str:
    """Construct name for the NXentry instances."""
    if len(parts) == 1:
        return align_name_part(parts[0])
    return "__".join([align_name_part(part) for part in parts])


def _format_value(value: int | float | str) -> int | float | str:
    """
    Formats the input value as an int or float if it's a numeric string.

    If a string represents a float (e.g., "5.0"), it remains a float even if it
    has no decimals.

    Args:
        value (Union[str, float, int]): The input value to format.

    Returns:
        Union[int, float, str]: The formatted value as int or float if numeric;
                                otherwise, returns the original value.

    """
    if isinstance(value, str) and re.match(r"^-?\d*\.?\d+(?:[eE][-+]?\d+)?$", value):
        # Check for decimal to ensure float, even if the decimal part is zero
        return float(value) if "." in value or "e" in value.lower() else int(value)
    return value


UNIT_PATTERN = re.compile(r"^([-+]?\d+(?:\.\d+)?(?:[eE][+-]?\d+)?)\s*([a-zA-Z/\s]+)$")


def split_value_and_unit(
    value_str: str,
) -> tuple[int | float | str, str]:
    """
    Splits a string into a numerical value and its associated unit.

    If the string contains a pattern where a number (which can be negative and
    in scientific notation) is followed by a unit, it returns a tuple containing
    the value (as a float) and the unit (as a string). If the string does not match
    this pattern, it returns the original string.

    Args:
        value_str (str): The input string to split.

    Returns:
        Tuple[Union[int, float, str], str]:
            - (value, unit) if a numeric value is detected.
            - (original string, "") if no numeric value is detected.

    """
    match = UNIT_PATTERN.match(value_str)
    if match:
        value = _format_value(match.group(1))
        unit = match.group(2).replace(" ", "") if match.group(2) else ""
        return value, unit
    return _format_value(value_str), ""


def extract_unit(
    key: str, value_str: str, unit_missing: dict[str, str] | None = None
) -> tuple[int | float | str, str]:
    """
    Extract a numeric value and its associated unit from a metadata string.

    The function identifies and separates numerical and unit components from
    the `value_str` string. If no unit is found in `value_str`, it checks the
    `unit_missing` dictionary for a default unit.

    Example:
        analyzer_work_function = "4.506eV"
        -> (4.506, "eV")

    Args:
        key (str): Key associated with the value.
        value_str (str): Combined numeric value and unit as a string.
        unit_missing (Optional[dict[str, str]]): Optional dictionary with default units
            for keys missing units. Defaults to None.

    Returns:
        tuple[Union[int, float, str], str]:
            - A tuple with the numeric value (int, float, or str) and the unit.
            - If no unit is found in `value_str`, it uses `unit_missing` if available.
            - If no unit is found in either, returns an empty string for the unit.

    """
    value, unit = split_value_and_unit(value_str)

    if not unit:
        unit = unit_missing.get(key, "")

    return value, unit


def check_units(template_path: str, unit: str) -> None:
    """
    Check that the unit is a valid pint unit.

    Args:
        template_path (str): Path of a Template object.
        unit (str): String representation of a unit.

    """
    error_txt = f"Invalid unit '{unit}' at path: {template_path}"

    if unit is not None:
        error_txt = f"Invalid unit '{unit}' at path: {template_path}"
        try:
            ureg.Unit(unit)
        except pint.errors.UndefinedUnitError as pint_err:
            raise pint.errors.UndefinedUnitError(error_txt) from pint_err
        except TypeError as type_err:
            logger.warning(f"WARNING: {error_txt}")



# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Data model for data from Scienta instruments.
"""

from dataclasses import dataclass, field

import numpy as np

from pynxtools_xps.reader_utils import XpsDataclass


@dataclass
class ScientaHeader(XpsDataclass):
    no_of_regions: int = 0
    software_version: str = ""


@dataclass
class ScientaRegion(XpsDataclass):
    region_id: int = 0
    region_name: str = ""
    energy_size: int = 0
    energy_scale: str = ""
    energy_scale_2: str = ""
    energy_axis: np.ndarray = field(default_factory=lambda: np.zeros(0))
    lens_mode: str = ""
    pass_energy: float = 0.0
    no_of_scans: int = 0
    excitation_energy: float = 0.0
    acquisition_mode: str = ""
    energy_units: str = "kinetic"  # ???
    center_energy: float = 0.0
    start_energy: float = 0.0
    stop_energy: float = 0.0
    step_size: float = 0.0
    dwell_time: float = 0.0
    detector_first_x_channel: int = 0
    detector_last_x_channel: int = 0
    detector_first_y_channel: int = 0
    detector_last_y_channel: int = 0
    number_of_slices: str = ""
    data_file: str = ""
    sequence_file: str = ""
    spectrum_type: str = ""
    instrument_name: str = ""
    vendor: str = ""
    user_name: str = ""
    sample_name: str = ""
    spectrum_comment: str = ""
    start_date: str = ""
    time: str = ""
    time_stamp: str = ""
    time_per_spectrum_channel: float = 0.0
    detector_mode: str = ""


"""
jsonschema data model for data from the Scienta PEAK software.
"""
scienta_igor_peak_schema = {
    "type": "object",
    "properties": {
        "Version": {"type": ["integer", "null"]},
        "Name": {"type": ["string", "null"]},
        "ElementSet": {"type": ["string", "null"]},
        "ImageSource": {"type": ["string", "null"]},
        "ImageDetector": {"type": ["string", "null"]},
        "LensMode": {"type": ["string", "null"]},
        "AcquisitionMode": {"type": ["string", "null"]},
        "PassEnergy": {"type": ["number", "null"]},
        "Date": {"type": ["string", "null"], "format": "date"},
        "Time": {"type": ["string", "null"], "format": "time"},
        "Duration": {"type": ["number", "null"]},
        "AcquisitionLoopCount": {"type": ["integer", "null"]},
        "Instrument": {"type": ["string", "null"]},
        "AnalyserModel": {"type": ["string", "null"]},
        "Location": {"type": ["string", "null"]},
        "SerialNumber": {"type": ["string", "null"]},
        "AnalyserRadius": {"type": ["number", "null"]},
        "AnalysedParticle": {"type": ["string", "null"]},
        "AnalyserWorkFunction": {"type": ["number", "null"]},
        "SpectrumVersion": {"type": ["integer", "null"]},
        "AcquisitionStatus": {"type": ["string", "null"]},
        "TotalAcquisitionTime": {"type": ["number", "null"]},
        "DwellTime": {"type": ["number", "null"]},
        "IsSweep": {"type": ["boolean", "null"]},
        "EnergyConversion": {
            "type": "object",
            "properties": {
                "UseAnalyserWorkFunction": {"type": ["boolean", "null"]},
                "UseSampleWorkFunction": {"type": ["boolean", "null"]},
                "UseSampleBias": {"type": ["boolean", "null"]},
                "AnalyserMode": {"type": ["string", "null"]},
            },
        },
        "EnergyMode": {"type": ["string", "null"]},
        "ConstantParameters": {"type": ["object", "null"]},
        "Slit": {
            "type": "object",
            "properties": {
                "Id": {"type": ["string", "null"]},
                "Name": {"type": ["string", "null"]},
                "SlitType": {"type": ["string", "null"]},
                "Width": {"type": ["number", "null"]},
                "Length": {"type": ["number", "null"]},
                "Radius": {"type": ["number", "null"]},
                "IsCentered": {"type": ["boolean", "null"]},
                "HasBadQuality": {"type": ["boolean", "null"]},
                "HasAperture": {"type": ["boolean", "null"]},
                "KnobPosition": {"type": ["integer", "null"]},
                "MotorizedPosition": {"type": ["integer", "null"]},
                "MotorizedPositionTolerance": {"type": ["number", "null"]},
            },
        },
        "Sample": {
            "type": "object",
            "properties": {
                "Name": {"type": ["string", "null"]},
                "Description": {"type": ["string", "null"]},
                "Temperature": {"type": ["number", "null"]},
                "Compound": {"type": ["string", "null"]},
                "Environment": {"type": ["string", "null"]},
                "WorkFunction": {"type": ["number", "null"]},
                "Bias": {"type": ["number", "null"]},
            },
        },
        "ExcitationSource": {
            "type": "object",
            "properties": {
                "ExcitationSourceServerInformation": {"type": ["string", "null"]},
                "ExcitationSourcePreset": {
                    "type": "object",
                    "properties": {
                        "Id": {"type": ["string", "null"]},
                        "Name": {"type": ["string", "null"]},
                        "Description": {"type": ["string", "null"]},
                        "TargetEnergy": {"type": ["number", "null"]},
                        "SourceType": {"type": ["string", "null"]},
                    },
                },
                "ExcitationSourceEnergyInformation": {
                    "type": "object",
                    "properties": {
                        "SourceType": {"type": ["string", "null"]},
                        "Energy": {"type": ["number", "null"]},
                        "TargetEnergy": {"type": ["number", "null"]},
                        "EnergyStatus": {"type": ["string", "null"]},
                        "EnergyStatusMessage": {"type": ["string", "null"]},
                    },
                },
                "ExcitationSourceServerStatusInformation": {"type": ["string", "null"]},
            },
        },
        "FixedAxes": {
            "type": "object",
            "patternProperties": {
                "^[XYZ]$": {
                    "type": "object",
                    "properties": {
                        "Center": {"type": ["number", "null"]},
                        "Binning": {"type": ["integer", "null"]},
                        "Name": {"type": ["string", "null"]},
                        "Unit": {"type": ["string", "null"]},
                    },
                }
            },
            "additionalProperties": True,
        },
        "SweepAxes": {
            "type": "object",
            "properties": {
                "X": {
                    "type": "object",
                    "properties": {
                        "Offset": {"type": ["number", "null"]},
                        "Delta": {"type": ["number", "null"]},
                        "Count": {"type": ["integer", "null"]},
                        "AnalyserDelta": {"type": ["number", "null"]},
                        "Name": {"type": ["string", "null"]},
                        "Unit": {"type": ["string", "null"]},
                    },
                }
            },
        },
        "UserName": {"type": ["string", "null"]},
        "StorageName": {"type": ["string", "null"]},
        "SequenceName": {"type": ["string", "null"]},
        "SequenceCounter": {"type": ["integer", "null"]},
        "SpectrumCounter": {"type": ["integer", "null"]},
        "TimePerEnergyChannel": {"type": ["number", "null"]},
    },
}
